# coding: utf8

"""
The 'clinica' executable command line, installed with the clinica package,
calls this module.

The aim of this module is to execute pipelines from the command line,
and gives to the user some other utils to work with the pipelines.
"""

import os
import sys

import argcomplete

from clinica.engine.cmdparser import *
from clinica.utils.stream import cprint

__author__ = "Michael Bacci"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Michael Bacci", "Alexandre Routier", "Mauricio Diaz"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Mauricio Diaz"
__email__ = "mauriciodiaz@inria.fr"
__status__ = "Development"


class ClinicaClassLoader:
    """
    Load pipelines from a custom locations (general from $HOME/clinica)
    """

    def __init__(self, env='CLINICAPATH',
                 baseclass=None, reg=r".*_cli\.py$", extra_dir=""):
        self.env = env
        if baseclass is None:
            import clinica.pipelines.engine as cpe
            self.baseclass = cpe.Pipeline
        else:
            self.baseclass = baseclass
        self.reg = reg
        self.extra_dir = extra_dir

    def load(self):
        import os
        pipeline_cli_parsers = []

        if self.env not in os.environ.keys():

            return pipeline_cli_parsers

        clinica_pipelines_path = join(os.environ[self.env], self.extra_dir)
        if not os.path.isdir(clinica_pipelines_path):
            return pipeline_cli_parsers

        src_path = self.discover_path_with_subdir(clinica_pipelines_path)
        self.add_to_python_path(src_path)
        files_match = self.find_files(src_path, self.reg)

        for file in files_match:
            pipeline_cli_parsers.append(self.load_class(self.baseclass, file))

        return pipeline_cli_parsers

    def load_class(self, baseclass, file):
        import imp
        import inspect
        py_module_name, ext = os.path.splitext(os.path.split(file)[-1])
        py_module = imp.load_source(py_module_name, file)
        for class_name, class_obj in inspect.getmembers(py_module, inspect.isclass):
            if inspect.isclass(class_obj) and not inspect.isabstract(class_obj):
                x = class_obj()
                if isinstance(x, baseclass):
                    return x

    def add_to_python_path(self, paths):
        for p in paths:
            if p not in sys.path:
                sys.path.append(p)

    def discover_path_with_subdir(self, path):
        return [os.path.join(path, file) for file in os.listdir(path) if os.path.isdir(os.path.join(path, file))]

    def find_files(self, paths, reg):
        import re
        return [os.path.join(path, file) for path in paths for file in os.listdir(path) if re.match(reg, file) is not None]


# Nice display
def custom_traceback(exc_type, exc_value, exc_traceback):
    import traceback
    import math
    from colorama import Fore
    from clinica.utils.exceptions import ClinicaException
    from clinica.utils.stream import cprint

    if issubclass(exc_type, ClinicaException):
        cprint(exc_value)
    elif issubclass(exc_type, KeyboardInterrupt):
        cprint('\n%s[Error] Program interrupted by the user. Clinica will now exit...%s' %
               (Fore.RED, Fore.RESET))
    else:
        cprint(Fore.RED + '\n' + '*' * 23 + '\n*** Clinica crashed ***\n' + '*' * 23 + '\n' + Fore.RESET)
        cprint('%sException type:%s %s' % (Fore.YELLOW, Fore.RESET, exc_type.__name__))
        cprint('%sException value:%s %s' % (Fore.YELLOW, Fore.RESET, exc_value))
        cprint('Below are displayed information that were gathered when Clinica crashed. This will help to understand '
               'what happened if you transfer those information to the Clinica development team.\n')

        frames = traceback.extract_tb(exc_traceback)
        framewidth = int(math.ceil(math.log(len(frames)) / math.log(10)))
        filewidth = 0
        linewidth = 0
        functionwidth = 0
        for frame in frames:
            filewidth = max(filewidth, len(frame[0]))
            linewidth = max(linewidth, frame[1])
            functionwidth = max(functionwidth, len(frame[2]))
        linewidth = int(math.ceil(math.log(linewidth) / math.log(10)))
        cprint('=' * (filewidth + linewidth + functionwidth + linewidth))
        for i in range(len(frames)):
            t = '{}' + ' ' * (1 + framewidth - len(str(i))) + Fore.RED \
                + '{}' + ' ' * (1 + filewidth - len(frames[i][0])) + Fore.RESET \
                + '{}' + ' ' * (1 + linewidth - len(str(frames[i][1]))) + Fore.GREEN \
                + '{}' + ' ' * (1 + functionwidth - len(frames[i][2])) + Fore.RESET + '{}'
            cprint(t.format(i, frames[i][0], frames[i][1], frames[i][2], frames[i][3]))
        cprint('=' * (filewidth + linewidth + functionwidth + linewidth))

    if not issubclass(exc_type, KeyboardInterrupt):
        cprint('\nDocumentation can be found here: %shttp://www.clinica.run/doc/%s\n'
               'If you need support, do not hesitate to ask: %shttps://groups.google.com/forum/#!forum/clinica-user%s\n'
               'Alternatively, you can also open an issue on GitHub: %shttps://github.com/aramis-lab/clinica/issues%s' %
               (Fore.BLUE, Fore.RESET, Fore.BLUE, Fore.RESET, Fore.BLUE, Fore.RESET))


def execute():
    import os
    import argparse
    from colorama import Fore
    import warnings
    import logging

    # Suppress potential warnings
    warnings.filterwarnings("ignore")

    # Remove warnings from duecredit package in order to silent
    # "Assuming non interactive session since isatty found missing" message
    logging.getLogger("duecredit.utils").setLevel(logging.ERROR)

    # Add warning message if PYTHONPATH is not empty
    # cf https://groups.google.com/forum/#!topic/clinica-user/bVgifEdkg20
    python_path = os.environ.get('PYTHONPATH', '')
    if python_path:
        print('%s[Warning] The PYTHONPATH environment variable is not empty.'
              ' Make sure there is no interference with Clinica (content of PYTHONPATH: %s).%s' %
              (Fore.YELLOW, python_path, Fore.RESET))

    # Nice traceback when clinica crashes
    sys.excepthook = custom_traceback

    MANDATORY_TITLE = (Fore.YELLOW + 'Mandatory arguments' + Fore.RESET)
    OPTIONAL_TITLE = (Fore.YELLOW + 'Optional arguments' + Fore.RESET)
    """
    Define and parse the command line argument
    """
    parser = ArgumentParser(add_help=False)
    parser.add_argument('-h', '--help', action='help',
                        default=argparse.SUPPRESS, help=argparse.SUPPRESS)
    parser._positionals.title = (
            Fore.YELLOW
            + 'clinica expects one of the following keywords'
            + Fore.RESET)
    parser._optionals.title = OPTIONAL_TITLE

    sub_parser = parser.add_subparsers(metavar='')
    parser.add_argument("-v", "--verbose",
                        dest='verbose',
                        action='store_true', default=False,
                        help='Verbose: print all messages to the console')
    parser.add_argument("-l", "--logname",
                        dest='logname',
                        default="clinica.log",
                        metavar=('file.log'),
                        help='Define the log file name (default: clinica.log)')

    """
    run category: run one of the available pipelines
    """
    from clinica.engine import CmdParser

    from clinica.pipelines.t1_freesurfer.t1_freesurfer_cli import T1FreeSurferCLI
    # from clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_longitudinal_cli import T1FreeSurferLongitudinalCLI
    from clinica.pipelines.t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_cli import T1VolumeTissueSegmentationCLI
    from clinica.pipelines.t1_volume_create_dartel.t1_volume_create_dartel_cli import T1VolumeCreateDartelCLI
    from clinica.pipelines.t1_volume_register_dartel.t1_volume_register_dartel_cli import T1VolumeRegisterDartelCLI
    from clinica.pipelines.t1_volume_dartel2mni.t1_volume_dartel2mni_cli import T1VolumeDartel2MNICLI
    from clinica.pipelines.t1_volume.t1_volume_cli import T1VolumeCLI
    from clinica.pipelines.t1_volume_existing_template.t1_volume_existing_template_cli import T1VolumeExistingTemplateCLI
    from clinica.pipelines.t1_volume_parcellation.t1_volume_parcellation_cli import T1VolumeParcellationCLI
    from clinica.pipelines.dwi_preprocessing_using_phasediff_fieldmap.dwi_preprocessing_using_phasediff_fieldmap_cli import DwiPreprocessingUsingPhaseDiffFieldmapCli
    from clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_cli import DwiPreprocessingUsingT1Cli
    from clinica.pipelines.dwi_dti.dwi_dti_cli import DwiDtiCli
    from clinica.pipelines.dwi_connectome.dwi_connectome_cli import DwiConnectomeCli
    from clinica.pipelines.fmri_preprocessing.fmri_preprocessing_cli import fMRIPreprocessingCLI
    from clinica.pipelines.pet_volume.pet_volume_cli import PETVolumeCLI
    from clinica.pipelines.pet_surface.pet_surface_cli import PetSurfaceCLI
    from clinica.pipelines.machine_learning_spatial_svm.spatial_svm_cli import SpatialSVMCLI
    from clinica.pipelines.statistics_surface.statistics_surface_cli import StatisticsSurfaceCLI
    pipelines = ClinicaClassLoader(baseclass=CmdParser,
                                   extra_dir="pipelines").load()
    pipelines += [
        T1FreeSurferCLI(),
        T1VolumeCLI(),
        # T1FreeSurferLongitudinalCLI(),
        DwiPreprocessingUsingPhaseDiffFieldmapCli(),
        DwiPreprocessingUsingT1Cli(),
        DwiDtiCli(),
        DwiConnectomeCli(),
        fMRIPreprocessingCLI(),
        PETVolumeCLI(),
        PetSurfaceCLI(),
        SpatialSVMCLI(),
        StatisticsSurfaceCLI(),
        T1VolumeExistingTemplateCLI(),
        T1VolumeTissueSegmentationCLI(),
        T1VolumeCreateDartelCLI(),
        T1VolumeRegisterDartelCLI(),
        T1VolumeDartel2MNICLI(),
        T1VolumeParcellationCLI()
    ]

    run_parser = sub_parser.add_parser(
        'run',
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter,
        help='To run pipelines on BIDS/CAPS datasets.'
    )
    run_parser.description = '%sRun pipelines on BIDS/CAPS datasets.%s' % \
                             (Fore.GREEN, Fore.RESET)
    run_parser._positionals.title = '%sclinica run expects one of the following pipelines%s' % \
                                    (Fore.GREEN, Fore.RESET)

    init_cmdparser_objects(
        parser,
        run_parser.add_subparsers(metavar='', dest='run'),
        pipelines
    )

    """
    convert category: convert one of the supported datasets into BIDS hierarchy
    """
    from clinica.iotools.converters.aibl_to_bids.aibl_to_bids_cli import AiblToBidsCLI
    from clinica.iotools.converters.adni_to_bids.adni_to_bids_cli import AdniToBidsCLI
    from clinica.iotools.converters.oasis_to_bids.oasis_to_bids_cli import OasisToBidsCLI
    from clinica.iotools.converters.nifd_to_bids.nifd_to_bids_cli import NifdToBidsCLI  # noga

    converters = ClinicaClassLoader(baseclass=CmdParser,
                                    extra_dir="iotools/converters").load()
    converters += [
        AdniToBidsCLI(),
        AiblToBidsCLI(),
        OasisToBidsCLI(),
        NifdToBidsCLI(),
    ]

    convert_parser = sub_parser.add_parser(
        'convert',
        add_help=False,
        help='To convert unorganized datasets into a BIDS hierarchy.',
    )
    convert_parser.description = '%sTools to convert unorganized datasets into a BIDS hierarchy.%s' % \
                                 (Fore.GREEN, Fore.RESET)
    convert_parser._positionals.title = '%sclinica convert expects one of the following datasets%s' % \
                                        (Fore.YELLOW, Fore.RESET)
    convert_parser._optionals.title = OPTIONAL_TITLE
    init_cmdparser_objects(
        parser,
        convert_parser.add_subparsers(metavar='', dest='convert'),
        converters
    )

    """
    iotools category
    """
    from clinica.iotools.utils.data_handling_cli import CmdParserSubjectsSessions
    from clinica.iotools.utils.data_handling_cli import CmdParserMergeTsv
    from clinica.iotools.utils.data_handling_cli import CmdParserMissingModalities
    from clinica.iotools.utils.data_handling_cli import CmdParserCenterNifti
    io_tools = [
        CmdParserSubjectsSessions(),
        CmdParserMergeTsv(),
        CmdParserMissingModalities(),
        CmdParserCenterNifti()
    ]

    HELP_IO_TOOLS = 'Tools to handle BIDS/CAPS datasets.'
    io_parser = sub_parser.add_parser(
        'iotools',
        add_help=False,
        help=HELP_IO_TOOLS,
    )
    io_parser.description = '%s%s%s' % (Fore.GREEN, HELP_IO_TOOLS, Fore.RESET)
    io_parser._positionals.title = '%sclinica iotools expects one of the following BIDS/CAPS utilities%s' % \
                                   (Fore.YELLOW, Fore.RESET)
    io_parser._optionals.title = OPTIONAL_TITLE

    init_cmdparser_objects(
        parser,
        io_parser.add_subparsers(metavar='', dest='iotools'),
        io_tools
    )

    """
    visualize category: run one of the available pipelines
    """
    from clinica.engine import CmdParser

    from clinica.pipelines.t1_freesurfer.t1_freesurfer_visualizer import T1FreeSurferVisualizer

    visualizers = ClinicaClassLoader(baseclass=CmdParser,
                                     extra_dir="pipelines").load()
    visualizers += [
        T1FreeSurferVisualizer(),
    ]

    visualize_parser = sub_parser.add_parser(
        'visualize',
        add_help=False,
        formatter_class=argparse.RawTextHelpFormatter,
        help='To visualize outputs of Clinica pipelines.'
    )
    visualize_parser.description = '%sVisualize outputs of Clinica pipelines.%s' % \
                                   (Fore.GREEN, Fore.RESET)
    visualize_parser._positionals.title = '%sclinica visualize expects one of the following pipelines%s' % \
                                          (Fore.YELLOW, Fore.RESET)

    init_cmdparser_objects(
        parser,
        visualize_parser.add_subparsers(metavar='', dest='visualize'),
        visualizers
    )

    """
    generate category: template
    """
    generate_parser = sub_parser.add_parser(
        'generate',
        add_help=False,
        help=('To generate pre-filled files when creating '
              'new pipelines (for developers).'),
    )
    generate_parser.description = '%sGenerate pre-filled files when creating new pipelines (for  developers).%s' % \
                                  (Fore.GREEN, Fore.RESET)
    generate_parser._positionals.title = '%sclinica generate expects one of the following tools%s' % \
                                         (Fore.YELLOW, Fore.RESET)
    generate_parser._optionals.title = OPTIONAL_TITLE

    from clinica.engine.template import CmdGenerateTemplates
    init_cmdparser_objects(
        parser,
        generate_parser.add_subparsers(metavar='', dest='generate'),
        [CmdGenerateTemplates()]
    )

    """
    Silent all sub-parser errors methods except the one which is called
    otherwise the output console will display useless messages
    """
    def silent_help(): pass

    def single_error_message(p):
        def error(x):
            from colorama import Fore
            print('%sError %s%s\n' % (Fore.RED, x, Fore.RESET))
            p.print_help()
            parser.print_help = silent_help
            exit(-1)
        return error
    for p in [run_parser, io_parser, convert_parser, generate_parser]:
        p.error = single_error_message(p)

    # Do not want stderr message
    def silent_msg(x):
        pass

    parser.error = silent_msg

    """
    Parse the command and check that everything went fine
    """
    args = None
    unknown_args = None
    try:
        argcomplete.autocomplete(parser)
        args, unknown_args = parser.parse_known_args()
        if unknown_args:
            if ('--verbose' in unknown_args) or ('-v' in unknown_args):
                cprint("Verbose detected")
                args.verbose = True
            unknown_args = [i for i in unknown_args if i != '-v']
            unknown_args = [i for i in unknown_args if i != '--verbose']
            if unknown_args:
                print('%s[Warning] Unknown flag(s) detected: %s. This will be ignored by Clinica%s' %
                      (Fore.YELLOW, unknown_args, Fore.RESET))
    except SystemExit:
        exit(0)
    except Exception:
        print("%s\n[Error] You wrote wrong arguments on the command line. Clinica will now exit.\n%s" % (Fore.RED, Fore.RESET))
        parser.print_help()
        exit(-1)

    if 'run' in args and hasattr(args, 'func') is False:
        # Case when we type `clinica run` on the terminal
        run_parser.print_help()
        exit(0)
    elif 'convert' in args and hasattr(args, 'func') is False:
        # Case when we type `clinica convert` on the terminal
        convert_parser.print_help()
        exit(0)
    elif 'iotools' in args and hasattr(args, 'func') is False:
        # Case when we type `clinica iotools` on the terminal
        io_parser.print_help()
        exit(0)
    elif 'visualize' in args and hasattr(args, 'func') is False:
        # Case when we type `clinica visualize` on the terminal
        visualize_parser.print_help()
        exit(0)
    elif 'generate' in args and hasattr(args, 'func') is False:
        # Case when we type `clinica generate` on the terminal
        generate_parser.print_help()
        exit(0)
    elif args is None or hasattr(args, 'func') is False:
        # Case when we type `clinica` on the terminal
        parser.print_help()
        exit(0)

    import clinica.utils.stream as var
    var.clinica_verbose = args.verbose

    if args.verbose is False:
        """
        Enable only cprint(msg) --> clinica print(msg)
        - All the print() will be ignored!
        - All the logging will be redirect to the log file.
        """
        from clinica.utils.stream import FilterOut
        sys.stdout = FilterOut(sys.stdout)
        import logging as python_logging
        from logging import Filter, ERROR
        import os
        from nipype import config, logging

        # Configure Nipype logger for our needs
        config.update_config({'logging': {'workflow_level': 'INFO',
                                          'log_directory': os.getcwd(),
                                          'log_to_file': True},
                              'execution': {'stop_on_first_crash': False,
                                            'hash_method': 'content'}
                              })
        logging.update_logging(config)

        # Define the LogFilter for ERROR detection
        class LogFilter(Filter):
            """
            The LogFilter class ables to monitor if an ERROR log signal is sent
            from Clinica/Nipype. If detected, the user will be warned.
            """
            def filter(self, record):
                if record.levelno >= ERROR:
                    import datetime
                    now = datetime.datetime.now().strftime('%H:%M:%S')
                    cprint(
                        '%s[%s]%s An error was found: please check the log file (%s) or crash file '
                        '(nipypecli crash <pklz_file>) for details. Clinica is still running.' %
                        (Fore.RED, now, Fore.RESET, args.logname))
                return True

        if args.verbose:
            logger = logging.getLogger('nipype.workflow')
            logger.addFilter(LogFilter())

        # Remove all handlers associated with the root logger object
        for handler in python_logging.root.handlers[:]:
            python_logging.root.removeHandler(handler)

        logging.disable_file_logging()

        # Enable file logging using a filename
        def enable_file_logging(self, filename):
            """
            Hack to define a filename for the log file! It overloads the
            'enable_file_logging' method in 'nipype/utils/logger.py' file.
            """
            import logging
            from logging.handlers import RotatingFileHandler as RFHandler
            config = self._config
            LOG_FILENAME = os.path.join(config.get('logging', 'log_directory'),
                                        filename)
            hdlr = RFHandler(LOG_FILENAME,
                             maxBytes=int(config.get('logging', 'log_size')),
                             backupCount=int(config.get('logging',
                                                        'log_rotate')))
            formatter = logging.Formatter(fmt=self.fmt, datefmt=self.datefmt)
            hdlr.setFormatter(formatter)
            self._logger.addHandler(hdlr)
            self._fmlogger.addHandler(hdlr)
            self._iflogger.addHandler(hdlr)
            self._hdlr = hdlr
        enable_file_logging(logging, args.logname)

        class Stream:
            def write(self, text):
                print(text)
                sys.stdout.flush()

        python_logging.basicConfig(
            format=logging.fmt, datefmt=logging.datefmt, stream=Stream())

    # Finally, run the command
    args.func(args)


if __name__ == '__main__':
    execute()



# coding: utf8

import clinica.engine as ce


class PETVolumeCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = 'pet-volume'

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = ('SPM-based pre-processing of PET images:\n'
                             'http://clinica.run/doc/Pipelines/PET_Volume/')

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # Clinica compulsory arguments (e.g. BIDS, CAPS, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("bids_directory",
                                  help='Path to the BIDS directory.')
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        clinica_comp.add_argument("group_id",
                                  help='User-defined identifier for the provided group of subjects.')
        # Optional arguments (e.g. FWHM)
        optional = self._args.add_argument_group(PIPELINE_CATEGORIES['OPTIONAL'])
        optional.add_argument("-psf", "--psf_tsv",
                              help='TSV file containing for each PET image its point spread function (PSF) measured '
                                   'at full-width and half maximum (fwhm_x, fwhm_y and fwhm_z).')
        optional.add_argument("-pet_tracer", "--pet_tracer", type=str, default='fdg', choices=['fdg', 'av45'],
                              help='PET tracer. Can be fdg or av45 (default: --pet_tracer fdg)')
        # Clinica standard arguments (e.g. --n_procs)
        self.add_clinica_standard_arguments()
        # Advanced arguments (i.e. tricky parameters)
        advanced = self._args.add_argument_group(PIPELINE_CATEGORIES['ADVANCED'])
        advanced.add_argument("-mask", "--mask_tissues", nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help="Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal "
                                   "fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to use "
                                   "for masking the PET image (default: GM, WM and CSF i.e. --mask_tissues 1 2 3).")
        advanced.add_argument("-threshold", "--mask_threshold", type=float, default=0.3,
                              help='Value used as threshold to binarize the tissue maps (default: --mask_threshold 0.3).')
        advanced.add_argument("-pvc_mask", "--pvc_mask_tissues",
                              nargs='+', type=int, default=[1, 2, 3], choices=range(1, 7),
                              help="Tissue classes (1: gray matter (GM), 2: white matter (WM), 3: cerebrospinal "
                                   "fluid (CSF), 4: bone, 5: soft-tissue, 6: background) to use "
                                   "as mask for PVC (default: GM, WM and CSF i.e. --pvc_mask_tissues 1 2 3).")
        advanced.add_argument("-smooth", "--smooth", nargs='+', type=int, default=[8], metavar="FWHM_N FWHM_M",
                              help="A list of integers (e.g. --smooth 6 8) specifying the different isomorphic FWHM "
                                   "in millimeters to smooth the image (default: --smooth 8).")
        list_atlases = ['AAL2', 'LPBA40', 'Neuromorphometrics', 'AICHA', 'Hammers']
        advanced.add_argument("-atlases", "--atlases",
                              nargs='+', type=str, metavar="",
                              default=list_atlases, choices=list_atlases,
                              help='A list of atlases used to calculate regional mean SUVRs '
                                   '(default: all atlases i.e. --atlases AAL2 AICHA Hammers LPBA40 Neuromorphometrics).')

    def run_command(self, args):
        """Run the pipeline with defined args."""
        from networkx import Graph
        from clinica.pipelines.pet_volume.pet_volume_pipeline import PETVolume
        from clinica.utils.check_dependency import verify_cat12_atlases
        from clinica.utils.ux import print_end_pipeline, print_crash_files_and_exit

        parameters = {
            'group_id': args.group_id,
            'psf_tsv': self.absolute_path(args.psf_tsv),
            'pet_tracer': args.pet_tracer,
            'mask_tissues': args.mask_tissues,
            'mask_threshold': args.mask_threshold,
            'pvc_mask_tissues': args.pvc_mask_tissues,
            'smooth': args.smooth,
            'atlases': args.atlases,
        }

        # If the user wants to use any of the atlases of CAT12 and has not installed it, we just remove it from the list
        # of the computed atlases
        parameters['atlases'] = verify_cat12_atlases(args.atlases)

        pipeline = PETVolume(
            bids_directory=self.absolute_path(args.bids_directory),
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory),
            parameters=parameters,
            name=self.name
        )

        if args.n_procs:
            exec_pipeline = pipeline.run(plugin='MultiProc',
                                         plugin_args={'n_procs': args.n_procs})
        else:
            exec_pipeline = pipeline.run()

        if isinstance(exec_pipeline, Graph):
            print_end_pipeline(self.name, pipeline.base_dir, pipeline.base_dir_was_specified)
        else:
            print_crash_files_and_exit(args.logname, pipeline.base_dir)



# coding: utf8

import clinica.pipelines.engine as cpe


class T1VolumeDartel2MNI(cpe.Pipeline):
    """T1VolumeDartel2MNI - Dartel template to MNI.

    Returns:
        A clinica pipeline object containing the T1VolumeDartel2MNI pipeline.
    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.group import check_group_label

        if 'group_id' not in self.parameters.keys():
            raise KeyError('Missing compulsory group_id key in pipeline parameter.')

        if 'tissues' not in self.parameters:
            self.parameters['tissues'] = [1, 2, 3]
        if 'voxel_size' not in self.parameters:
            self.parameters['voxel_size'] = None
        if 'modulate' not in self.parameters:
            self.parameters['modulate'] = True
        if 'smooth' not in self.parameters:
            self.parameters['smooth'] = [8]

        check_group_label(self.parameters['group_id'])

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """

        return ['native_segmentations', 'flowfield_files', 'template_file']

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """

        return ['normalized_files', 'smoothed_normalized_files', 'atlas_statistics']

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os
        from colorama import Fore
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.inputs import clinica_file_reader, clinica_group_reader
        from clinica.utils.input_files import (t1_volume_final_group_template,
                                               t1_volume_native_tpm,
                                               t1_volume_deformation_to_template)
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_groups_in_caps_directory, print_images_to_process

        # Check that group already exists
        if not os.path.exists(os.path.join(self.caps_directory, 'groups', 'group-' + self.parameters['group_id'])):
            print_groups_in_caps_directory(self.caps_directory)
            raise ClinicaException(
                '%sGroup %s does not exist. Did you run t1-volume or t1-volume-create-dartel pipeline?%s' %
                (Fore.RED, self.parameters['group_id'], Fore.RESET)
            )

        all_errors = []
        read_input_node = npe.Node(name="LoadingCLIArguments",
                                   interface=nutil.IdentityInterface(
                                       fields=self.get_input_fields(),
                                       mandatory_inputs=True))

        # Segmented Tissues
        # =================
        tissues_input = []
        for tissue_number in self.parameters['tissues']:
            try:
                native_space_tpm = clinica_file_reader(
                    self.subjects,
                    self.sessions,
                    self.caps_directory,
                    t1_volume_native_tpm(tissue_number))
                tissues_input.append(native_space_tpm)
            except ClinicaException as e:
                all_errors.append(e)
        # Tissues_input has a length of len(self.parameters['mask_tissues']). Each of these elements has a size of
        # len(self.subjects). We want the opposite : a list of size len(self.subjects) whose elements have a size of
        # len(self.parameters['mask_tissues']. The trick is to iter on elements with zip(*my_list)
        tissues_input_rearranged = []
        for subject_tissue_list in zip(*tissues_input):
            tissues_input_rearranged.append(subject_tissue_list)

        read_input_node.inputs.native_segmentations = tissues_input_rearranged

        # Flow Fields
        # ===========
        try:
            read_input_node.inputs.flowfield_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                t1_volume_deformation_to_template(self.parameters['group_id']))
        except ClinicaException as e:
            all_errors.append(e)

        # Dartel Template
        # ================
        try:
            read_input_node.inputs.template_file = clinica_group_reader(
                self.caps_directory,
                t1_volume_final_group_template(self.parameters['group_id'])
            )
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = 'Clinica faced error(s) while trying to read files in your CAPS/BIDS directories.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaCAPSError(error_message)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint('The pipeline will last a few minutes per image.')

        self.connect([
            (read_input_node, self.input_node, [('native_segmentations', 'native_segmentations')]),
            (read_input_node, self.input_node, [('flowfield_files', 'flowfield_files')]),
            (read_input_node, self.input_node, [('template_file', 'template_file')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.pipeline.engine as npe
        import nipype.interfaces.io as nio
        from clinica.utils.filemanip import zip_nii

        # Writing normalized images (and smoothed) into CAPS
        # ==================================================
        write_normalized_node = npe.MapNode(name='write_normalized_node',
                                            iterfield=['container', 'normalized_files', 'smoothed_normalized_files'],
                                            interface=nio.DataSink(infields=['normalized_files',
                                                                             'smoothed_normalized_files']))
        write_normalized_node.inputs.base_directory = self.caps_directory
        write_normalized_node.inputs.parameterization = False
        write_normalized_node.inputs.container = ['subjects/' + self.subjects[i] + '/' + self.sessions[i] +
                                                  '/t1/spm/dartel/group-' + self.parameters['group_id']
                                                  for i in range(len(self.subjects))]
        write_normalized_node.inputs.regexp_substitutions = [
            (r'(.*)c1(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-graymatter_probability\3'),
            (r'(.*)c2(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-whitematter_probability\3'),
            (r'(.*)c3(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-csf_probability\3'),
            (r'(.*)c4(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-bone_probability\3'),
            (r'(.*)c5(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-softtissue_probability\3'),
            (r'(.*)c6(sub-.*)(\.nii(\.gz)?)$', r'\1\2_segm-background_probability\3'),
            (r'(.*)mw(sub-.*)_probability(\.nii(\.gz)?)$', r'\1\2_space-Ixi549Space_modulated-on_probability\3'),
            (r'(.*)w(sub-.*)_probability(\.nii(\.gz)?)$', r'\1\2_space-Ixi549Space_modulated-off_probability\3'),
            (r'(.*)/normalized_files/(sub-.*)$', r'\1/\2'),
            (r'(.*)/smoothed_normalized_files/(fwhm-[0-9]+mm)_(sub-.*)_probability(\.nii(\.gz)?)$',
             r'\1/\3_\2_probability\4'),
            (r'trait_added', r'')
        ]
        self.connect([
            (self.output_node, write_normalized_node, [(('normalized_files', zip_nii, True), 'normalized_files'),
                                                       (('smoothed_normalized_files', zip_nii, True),
                                                        'smoothed_normalized_files')])
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.spm as spm
        import nipype.pipeline.engine as npe
        import nipype.interfaces.utility as nutil
        from clinica.utils.filemanip import unzip_nii
        from ..t1_volume_dartel2mni import t1_volume_dartel2mni_utils as dartel2mni_utils

        # Unzipping
        # =========
        unzip_tissues_node = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                        output_names=['out_file'],
                                                        function=unzip_nii),
                                         name='unzip_tissues_node',
                                         iterfield=['in_file'])
        unzip_flowfields_node = npe.MapNode(nutil.Function(input_names=['in_file'],
                                                           output_names=['out_file'],
                                                           function=unzip_nii),
                                            name='unzip_flowfields_node',
                                            iterfield=['in_file'])
        unzip_template_node = npe.Node(nutil.Function(input_names=['in_file'],
                                                      output_names=['out_file'],
                                                      function=unzip_nii),
                                       name='unzip_template_node')

        # DARTEL2MNI Registration
        # =======================
        dartel2mni_node = npe.MapNode(spm.DARTELNorm2MNI(),
                                      name='dartel2MNI',
                                      iterfield=['apply_to_files', 'flowfield_files'])
        if self.parameters['voxel_size'] is not None:
            dartel2mni_node.inputs.voxel_size = tuple(self.parameters['voxel_size'])
        dartel2mni_node.inputs.modulate = self.parameters['modulate']
        dartel2mni_node.inputs.fwhm = 0

        # Smoothing
        # =========
        if self.parameters['smooth'] is not None and len(self.parameters['smooth']) > 0:
            smoothing_node = npe.MapNode(spm.Smooth(),
                                         name='smoothing_node',
                                         iterfield=['in_files'])

            smoothing_node.iterables = [('fwhm', [[x, x, x] for x in self.parameters['smooth']]),
                                        ('out_prefix', ['fwhm-' + str(x) + 'mm_' for x in self.parameters['smooth']])]
            smoothing_node.synchronize = True

            join_smoothing_node = npe.JoinNode(interface=nutil.Function(input_names=['smoothed_normalized_files'],
                                                                        output_names=['smoothed_normalized_files'],
                                                                        function=dartel2mni_utils.join_smoothed_files),
                                               joinsource='smoothing_node',
                                               joinfield='smoothed_normalized_files',
                                               name='join_smoothing_node')
            self.connect([
                (dartel2mni_node, smoothing_node, [('normalized_files', 'in_files')]),
                (smoothing_node, join_smoothing_node, [('smoothed_files', 'smoothed_normalized_files')]),
                (join_smoothing_node, self.output_node, [('smoothed_normalized_files', 'smoothed_normalized_files')])
            ])
        else:
            self.output_node.inputs.smoothed_normalized_files = []

        # Connection
        # ==========
        self.connect([
            (self.input_node, unzip_tissues_node, [('native_segmentations', 'in_file')]),
            (self.input_node, unzip_flowfields_node, [('flowfield_files', 'in_file')]),
            (self.input_node, unzip_template_node, [('template_file', 'in_file')]),
            (unzip_tissues_node, dartel2mni_node, [('out_file', 'apply_to_files')]),
            (unzip_flowfields_node, dartel2mni_node, [
                (('out_file', dartel2mni_utils.prepare_flowfields, self.parameters['tissues']), 'flowfield_files')
            ]),
            (unzip_template_node, dartel2mni_node, [('out_file', 'template_file')]),
            (dartel2mni_node, self.output_node, [('normalized_files', 'normalized_files')])
        ])



# coding: utf8


def atlas_decomposition(dartel_input):
    """

    :param dartel_input: path to the dartel input
    :return: 3 atlases (gm, wm, csf)
    """

    import nibabel as nib

    dartel = nib.load(dartel_input)
    dartel = dartel.get_data()
    atlas_1 = dartel[:, :, :, 0]
    atlas_2 = dartel[:, :, :, 1]
    atlas_3 = dartel[:, :, :, 2]
    atlas = [atlas_1, atlas_2, atlas_3]

    return atlas


def spm_read(fname):
    """
     Read the image and the header of fname
    :param fname: name of the image
    :return: it returns the image data as an array nibabel and the header
    """

    import nibabel as nib

    import numpy as np

    img = nib.load(fname)
    pico = img.get_data()
    pico = np.array(pico, dtype='float32')
    mask = np.isnan(pico)
    pico[mask] = 0
    volu = img.header

    return [pico, volu]


def spm_write_vol(fname, regularized_features):
    """

    :param fname: name of the atlas image, necessary to take information about the affine matrix and the header
    :param regularized_features: new image data obtained with the regularization of the input image
    :return: new image data in NIFTI format
    """

    import nibabel as nib

    i = nib.load(fname)
    data = regularized_features
    img = nib.Nifti1Image(data, header=i.header, affine=i.affine)
    return img


def rescaleImage(image1, p):
    """
    Normalization of the histogram of intensity
    :param Image1: input image
    :param p: vector of minimum and maximum value for the normalization
    how the istogram is normalized:
    between [0 1] if there are no options
    beween [1 p] if len(p) ==1
    between [p[0] p[1]] if len(p) == 2
    :return: image with the histogram normalized

    """
    import numpy as np
    eps = 2.2204e-16
    p = np.array(p)
    m = image1.min()
    M = (image1 - m).max()

    image2 = (image1 - m) / (M + eps)

    if len(p) == 1:
        image2 = image2 * (p - 1) + 1
    elif len(p) == 2:
        image2 = image2 * (p[1] - p[0]) + p[0]

    return image2


def tensor_scalar_product(sc, g1):
    """

    :param sc: scalar
    :param g1: 3 * 3 tensor
    :return: product between the tensor and the scalar
    """
    import numpy as np
    # we define the scalar and the tensor as complex

    sc = np.array(sc, dtype=np.complex128)
    g1 = np.array(g1, dtype=np.complex128)
    g = np.zeros(g1.shape, dtype=np.complex128)  # new vector

    for i in range(g1.shape[0]):
        for j in range(g1.shape[1]):
            for k in range(g1.shape[0]):
                g[i][j] = g1[i][j] * sc

    # g is the final tensor
    return g


def tensor_eye(atlas):
    """

    :param atlas: list of atlases
    :return: the identity matrix of a tensor
    """
    import numpy as np
    a = np.ones(atlas[0].shape)
    b = np.zeros(atlas[0].shape)
    # the new matrix is 1 on the diagonal and 0 in the other positions
    X = [a, b, b]
    Y = [b, a, b]
    Z = [b, b, a]

    matrix = [X, Y, Z]
    # it returns matrix, combination of array
    return matrix


def tensor_sum(g1, g2):
    """

    :param g1: first tensor
    :param g2: second tensor
    :return: sum of the two tensor
    """
    import numpy as np
    g = np.add(g1, g2)
    # numpy add to sum the tensors
    return g


def tensor_product(g1, g2):
    """

    :param g1: first tensor
    :param g2: second tensor
    :return: product between the two tensors
    """
    import numpy as np
    g1 = np.array(g1)
    g2 = np.array(g2)

    g = np.zeros(g1.shape)

    for i in range(g1.shape[0]):
        for j in range(g2.shape[0]):
            g[i][j] = 0
            for k in range(g1.shape[0]):
                g[i][j] = g[i][j] + np.multiply(g1[i][k], g2[k][j])

    # g = g1 * g2 (dim of the tensor: 3*3*xg*yg*zg)
    return g


def tensor_determinant(g):
    """

    :param g: tensor dim = 3*3*xg*yg*z
    :return: determinant of the tensor dim = xg*yg*zg
    """
    import numpy as np
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    g = np.array(g)
    d = 0
    s = g.shape
    # recursive function we have a matrix of 3*3 and then we divide it in different blocks of 2*2 and we calculate the determinant
    # of them making a um of the different determinants. The resulting determinant it's a sum of the different blocks.
    if s[0] == 3:
        # if the tensor is 3*3

        for i in range(s[0]):

            if np.mod(i, 2) == 0:
                epsilon = 1
            else:
                epsilon = -1

            if i == 0:
                g1 = [[g[1][1], g[1][2]], [g[2][1], g[2][2]]]
            elif i == 1:
                g1 = [[g[0][1], g[0][2]], [g[2][1], g[2][2]]]
            else:
                g1 = [[g[0][1], g[0][2]], [g[1][1], g[1][2]]]
            # it's a recursive function
            prod = epsilon * g[i][0] * utils.tensor_determinant(g1)
            d = d + prod

    elif s[0] == 2:
        # if the tensor is 2*2
        for i in range(s[0]):
            if np.mod(i, 2) == 0:
                epsilon = 1
            else:
                epsilon = -1
            if i == 0:
                g1 = [g[1][1]]
            elif i == 1:
                g1 = [g[0][1]]
            prod = epsilon * g[i][0] * utils.tensor_determinant(g1)
            d = d + prod

    elif s[0] == 1:
        # if the tensor is 1 matrix
        d = [g[0]]
    return d


def tensor_trace(g):
    """

    :param g: tensor
    :return: trace of a tensor
    """
    import numpy as np
    trace = np.trace(g)
    # trace if the trace of the input tensor
    return trace


def roots_poly(C):
    """

    :param C: coefficients. If C has N+1 components, the polynomial is C(1)*X^N + ... + C(N) * X + C(N+1)
    :return: roots of the polynomial

    the functions find the polynomial roots. It computes the roots of the polynomail whose coefficients are the elements of the vector C?
    """
    import numpy as np
    import math

    import cmath
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    C = np.array(C)
    if C.shape[0] < 2:
        rts = []

    elif C.shape[0] < 3:
        rts = - C[:, 1] * (1 / C[:, 0])

    elif C.shape[0] < 4:
        # implementation of delta
        delta = np.array([cmath.sqrt((C[1, i] * C[1, i]) - (4 * C[0, i] * C[2, i])) for i in range(C.shape[1])])
        # two roots
        rts1 = (-C[1, :] + delta) * (1 / ((2 * C[0, :])))
        rts2 = (-C[1, :] - delta) * (1 / ((2 * C[0, :])))
        rts = np.array([rts1, rts2])

    elif C.shape[0] < 5:
        # implementation of the method of Cardan

        a = C[0, :]
        b = C[1, :]
        c = C[2, :]
        d = C[3, :]

        p = - b * b * (1 / (3 * a * a)) + c * (1 / a)
        q = b * (1 / (27. * a)) * (2 * b * b * (1 / (a * a)) - 9 * c * (1 / a)) + d * (1 / a)

        new_roots = np.array([np.ones((q.shape[0])), q, - (p * p * p) / 27])

        rts = utils.roots_poly(new_roots)

        u = rts[0, :]
        u_mod = abs(u) ** (1 / 3)
        u_angle = np.angle(u) * (1 / 3)

        v = rts[1, :]
        v_mod = abs(v) ** (1 / 3)
        v_angle = np.angle(v) * (1 / 3)

        rts = np.zeros([C.shape[1], 3], dtype='complex128')

        ind = np.zeros([u.shape[0]], dtype='int32')

        for k in [0, 1, 2]:
            u = u_mod * np.power(math.e, 1j * (u_angle + k * 2 * math.pi * (1 / 3)))
            u = np.array(u)

            for r in [0, 1, 2]:
                v = v_mod * np.power(math.e, 1j * (v_angle + r * 2 * math.pi * (1 / 3)))

                ind2 = abs(u * v + p * 1 / 3) < 1e-10

                rts[ind2, ind[ind2]] = u[ind2] + v[ind2] - b[ind2] * (1 / (3 * a[ind2]))

                ind[ind2] = np.minimum(2, ind[ind2] + 1)

    else:
        print("For degree > 3 use roots ")

    return rts


def tensor_eigenvalues(g):
    """

    :param g: tensor
    :return: eigenvalues of the tensor

    """
    import numpy as np
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    g = np.array(g)

    if g.shape[0] < 4:
        # condition if we have a tensor

        C1 = np.ones(len(np.ravel(g[0][0])))
        buff = -utils.tensor_trace(g)
        C2 = buff.flatten('F')
        buff = utils.tensor_trace(utils.tensor_product(g, g))
        buff = (- 0.5 * (buff.flatten('F') - np.multiply(C2, C2)))
        C3 = buff.flatten('F')
        buff = -utils.tensor_determinant(g)
        C4 = buff.flatten('F')

        C = np.array([C1, C2, C3, C4])
        rts = utils.roots_poly(C)

    else:
        print('Degree too big : not still implemented')

    rts2 = rts.real.copy()
    rts2.sort()

    lamb = np.zeros(shape=(g.shape[0], g.shape[2], g.shape[3], g.shape[4]), dtype='complex128')

    for i in range(g.shape[0]):
        lamb[i, :, :, :] = rts2[:, i].reshape(g.shape[2], g.shape[3], g.shape[4], order='F')

    # lamb[0] is the smallest eigenvalues, lamb[2] is the biggest
    return lamb


def tensor_transpose(g):
    """

    :param g: tensor
    :return: tranpose of the tensor
    """
    import numpy as np

    g = np.array(g)
    tg = np.array(g)
    for i in range(g.shape[0]):
        for j in range(g.shape[0]):
            tg[i][j] = g[j][i]
    # tg is the transposed tensor
    return tg


def tensor_commatrix(g):
    """

    :param g: tensor
    :return: commatrix of the tensor
    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    import numpy as np
    g = np.array(g)
    g_com = []

    for i in range(g.shape[0]):
        for j in range(g.shape[0]):
            if np.mod(i + j, 2) == 0:
                epsilon = 1
            else:
                epsilon = -1
            if i == 0:
                if j == 0:
                    n = [[g[1][1], g[1][2]], [g[2][1], g[2][2]]]
                    a0 = epsilon * utils.tensor_determinant(n)
                elif j == 1:
                    n = [[g[1][0], g[1][2]], [g[2][0], g[2][2]]]
                    a1 = epsilon * utils.tensor_determinant(n)
                else:
                    n = [[g[1][0], g[1][1]], [g[2][0], g[2][1]]]
                    a2 = epsilon * utils.tensor_determinant(n)
            elif i == 1:
                if j == 0:
                    n = [[g[0][1], g[0][2]], [g[2][1], g[2][2]]]
                    b0 = epsilon * utils.tensor_determinant(n)
                elif j == 1:
                    n = [[g[0][0], g[0][2]], [g[2][0], g[2][2]]]
                    b1 = epsilon * utils.tensor_determinant(n)
                else:
                    n = [[g[0][0], g[0][1]], [g[2][0], g[2][1]]]
                    b2 = epsilon * utils.tensor_determinant(n)
            else:
                if j == 0:
                    n = [[g[0][1], g[0][2]], [g[1][1], g[1][2]]]
                    c0 = epsilon * utils.tensor_determinant(n)
                elif j == 1:
                    n = [[g[0][0], g[0][2]], [g[1][0], g[1][2]]]
                    c1 = epsilon * utils.tensor_determinant(n)
                else:
                    n = [[g[0][0], g[0][1]], [g[1][0], g[1][1]]]
                    c2 = epsilon * utils.tensor_determinant(n)
    g_com = [[a0, a1, a2], [b0, b1, b2], [c0, c1, c2]]
    g_com = np.array(g_com)
    if len(g_com.shape) == 6:
        g_com = g_com[:, :, 0, :, :, :]

    return g_com


def create_fisher_tensor(atlas):
    """

    :param atlas: list of 3 atlases, the 3 probability maps from the template with 3 components
    :return: g = tensor
    """

    # create tensor for fisher metrics
    import numpy as np
    import math

    upper_bound = 0.999  # probabibilty limits to avoid log(0) and log(1)
    lower_bound = 0.001  # probability limits to avoid log(0) and log(1)
    epsilon = 1e-6  # regularization

    n = atlas[0].shape

    a = np.ones(atlas[0].shape)
    b = np.zeros(atlas[0].shape)
    X = [np.dot(a, epsilon), b, b]
    Y = [b, np.dot(a, epsilon), b]
    Z = [b, b, np.dot(a, epsilon)]

    g = [X, Y, Z]

    for i in range(3):  # for for each component of the tensor
        proba = 1
        proba = (proba * atlas[i])
        proba = np.maximum(np.minimum(proba, upper_bound), lower_bound)
        gr = np.array(np.gradient(np.log(proba)))

        for x in range(3):
            for y in range(3):
                g[x][y] = (g[x][y] + (proba * gr[x] * gr[y]))

    return g


def tensor_helmholtz(x, h, detg, k):
    """

    :param x: 3D Array
    :param h: sqrt(det(g)) * inverse(g)^2 -> g is he metric tensor of the 3D manifold M
    :param detg: sqrt(det(g))
    :param k: constant( 0 - gives laplacian)
    :return:

    """

    import numpy as np
    detg = np.array(detg)
    if len(detg.shape) == 3:
        detg_ = detg
    else:
        detg_ = detg[0]

    weight = detg_[1:-1, 1:-1, 1:-1] * k
    if len(h.shape) == 6:
        h_ = h[:, :, 0, :, :, :, ]
    else:
        h_ = h
    for i in range(len(h)):  # from 1 to 3
        mat = h_[i][i]
        weight = weight + mat[1:-1, 1:-1, 1:-1]

    mat_1 = h_[0][0]
    weight = weight + 0.5 * mat_1[:-2, 1:-1, 1:-1]
    weight = weight + 0.5 * mat_1[2:, 1:-1, 1:-1]

    mat_2 = h_[1][1]
    weight = weight + 0.5 * mat_2[1:-1, :-2, 1:-1]
    weight = weight + 0.5 * mat_2[1:-1, 2:, 1:-1]

    mat_3 = h_[2][2]
    weight = weight + 0.5 * mat_3[1:-1, 1:-1, :-2]
    weight = weight + 0.5 * mat_3[1:-1, 1:-1, 2:]

    y0 = (weight * x[1:-1, 1:-1, 1:-1])

    mat1 = h_[0][1]
    mat2 = h_[1][0]
    mat3 = h_[1][2]
    mat4 = h_[2][1]
    mat5 = h_[0][2]
    mat6 = h_[2][0]

    y0 = y0 + (-1 * -1 * -0.25) * (mat1[:-2, 1:-1, 1: -1] + mat2[1:-1, :-2, 1:-1]) * x[:-2, :-2, 1:-1]
    y0 = y0 + (-1 * -1 * -0.25) * (mat3[1:-1, :-2, 1: -1] + mat4[1:-1, 1:-1, :-2]) * x[1:-1, :-2, :-2]
    y0 = y0 + (-1 * -1 * -0.25) * (mat5[:-2, 1:-1, 1: -1] + mat6[1:-1, 1:-1, :-2]) * x[:-2, 1:-1, :-2]
    y0 = y0 + (-1 * +1 * -0.25) * (mat1[:-2, 1:-1, 1: -1] + mat2[1:-1, 2:, 1:-1]) * x[:-2, 2:, 1:-1]
    y0 = y0 + (-1 * +1 * -0.25) * (mat3[1:-1, :-2, 1: -1] + mat4[1:-1, 1:-1, 2:]) * x[1:-1, :-2, 2:]
    y0 = y0 + (-1 * +1 * -0.25) * (mat5[:-2, 1:-1, 1: -1] + mat6[1:-1, 1:-1, 2:]) * x[:-2, 1:-1, 2:]
    y0 = y0 + (+1 * -1 * -0.25) * (mat1[2:, 1:-1, 1: -1] + mat2[1:-1, :-2, 1:-1]) * x[2:, :-2, 1:-1]
    y0 = y0 + (+1 * -1 * -0.25) * (mat3[1:-1, 2:, 1: -1] + mat4[1:-1, 1:-1, :-2]) * x[1:-1, 2:, :-2]
    y0 = y0 + (+1 * -1 * -0.25) * (mat5[2:, 1:-1, 1: -1] + mat6[1:-1, 1:-1, :-2]) * x[2:, 1:-1, :-2]
    y0 = y0 + (+1 * +1 * -0.25) * (mat1[2:, 1:-1, 1: -1] + mat2[1:-1, 2:, 1:-1]) * x[2:, 2:, 1:-1]
    y0 = y0 + (+1 * +1 * -0.25) * (mat3[1:-1, 2:, 1: -1] + mat4[1:-1, 1:-1, 2:]) * x[1:-1, 2:, 2:]
    y0 = y0 + (+1 * +1 * -0.25) * (mat5[2:, 1:-1, 1: -1] + mat6[1:-1, 1:-1, 2:]) * x[2:, 1:-1, 2:]

    y0 = y0 + (-0.5) * (mat_1[1:-1, 1:-1, 1:-1] + mat_1[:-2, 1:-1, 1:-1]) * x[:-2, 1:-1, 1:-1]
    y0 = y0 + (-0.5) * (mat_1[1:-1, 1:-1, 1:-1] + mat_1[2:, 1:-1, 1:-1]) * x[2:, 1:-1, 1:-1]
    y0 = y0 + (-0.5) * (mat_2[1:-1, 1:-1, 1:-1] + mat_2[1:-1, :-2, 1:-1]) * x[1:-1, :-2, 1:-1]
    y0 = y0 + (-0.5) * (mat_2[1:-1, 1:-1, 1:-1] + mat_2[1:-1, 2:, 1:-1]) * x[1:-1, 2:, 1:-1]
    y0 = y0 + (-0.5) * (mat_3[1:-1, 1:-1, 1:-1] + mat_3[1:-1, 1:-1, :-2]) * x[1:-1, 1:-1, :-2]
    y0 = y0 + (-0.5) * (mat_3[1:-1, 1:-1, 1:-1] + mat_3[1:-1, 1:-1, 2:]) * x[1:-1, 1:-1, 2:]

    return y0


def tensor_inverse(g):
    """

    :param g: tensor
    :return: inverse of the tensor
    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    import numpy as np

    h = utils.tensor_transpose(utils.tensor_commatrix(g))
    detg = utils.tensor_determinant(g)

    h = h * (1 / (detg))
    mask = np.isnan(h)
    h[mask] = 0
    return h


def operateur(x, ginv, detg):
    """

    :param x:
    :param ginv:
    :param detg:
    :return:
    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    import numpy as np

    if len(x.shape) == 4:
        x = x[0, :, :, :]
    y = np.zeros([x.shape[0] + 2, x.shape[1] + 2, x.shape[2] + 2])
    y = np.array(y, dtype=np.complex_)
    y[1:-1, 1:-1, 1:-1] = x
    y = utils.tensor_helmholtz(y, ginv, detg, 0)

    return y


def largest_eigenvalue_heat_3D_tensor2(g, h, epsilon):
    """

    :param g: metric tensor
    :param h: space step
    :param epsilon: stop criterion
    :return: lamba = the largest eigenvalues

    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    import numpy as np
    import cmath

    # parameters
    if epsilon is None:
        epsilon = 1e-6
    erreur = 1 + epsilon

    # tensors

    detg = utils.tensor_determinant(g)
    detg = np.array(detg, dtype=np.complex128)  # complex tensor
    detg = np.sqrt(detg)
    detg = detg[0]
    ginv = utils.tensor_inverse(g)

    if len(ginv.shape) == 6:
        ginv = ginv[:, :, 0, :, :, :]

    ginv = utils.tensor_scalar_product(detg, ginv)
    detg2 = detg[1:-1, 1:-1, 1:-1]  # 141*121*141
    detg2[np.isnan(detg2)] = 0
    detg[np.isnan(detg)] = 0
    ginv[np.isnan(ginv)] = 0

    # initialisation

    s = [g[0][0].shape[0] - 2, g[0][0].shape[1] - 2, g[0][0].shape[2] - 2]
    b1 = np.ones([s[0], s[1], s[2]])

    b1 = np.divide(b1, np.array(cmath.sqrt(np.dot(b1.flatten('F').transpose(), b1.flatten('F'))), dtype=np.complex128))

    print("Computation of the largest eigenvalue ...")
    while erreur > epsilon:
        b0 = b1
        b2 = np.array(np.divide(np.array(utils.operateur(b1, ginv, detg)) * h, detg2) / h / h / h, dtype=np.complex128)
        b1 = np.divide(b2, np.array(cmath.sqrt(np.dot(b2.flatten('F').transpose(), b2.flatten('F')))),
                       dtype=np.complex128)

        erreur = np.linalg.norm(b1.flatten('F') - b0.flatten('F'))

    print("done")

    lam = (cmath.sqrt(np.dot(b2.flatten('F').transpose(), b2.flatten('F'))))

    return lam


def heat_finite_elt_3D_tensor2(x0, t_final, t_step, h, g):
    """

    :param x0: vector x (at t = 0)
    :param t_final: time
    :param t_step: time step (must satisfy the CFL max(lambda) < 2)
    :param h:
    :param g: metric tensor
    :return: vector x (at t = t_final)

    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    import numpy as np
    if len(x0.shape) == 4:
        x0 = x0[0, :, :, :]

    # parameters
    nb_step = np.ceil(t_final / t_step)  # number of time step
    nb_step = nb_step.astype(int)
    t_step = t_final / nb_step

    # tensors
    detg = utils.tensor_determinant(g)
    detg = np.sqrt(detg)
    ginv = utils.tensor_inverse(g)
    ginv = utils.tensor_scalar_product(detg, ginv)
    detg2 = detg[:, 1:-1, 1:-1, 1:-1]
    if len(ginv.shape) == 6:
        ginv = ginv[:, :, 0, :, :, :]
    if len(detg.shape) == 4:
        detg = detg[0, :, :, :]
    ginv = np.array(ginv.real, dtype='float64')
    detg = np.array(detg.real, dtype='float64')
    detg2 = np.array(detg2.real, dtype='float64')

    # LOOP
    x = x0
    for i in range(nb_step):
        x = np.array(x - t_step * (np.divide(np.array(utils.operateur(x, ginv, detg)) * h, detg2)) / h / h / h)

    return x


def heat_finite_elt_2D_tensor2(x0, t_final, t_step, h, g):
    """

    :param x0: vector x (at t = 0)
    :param t_final: time
    :param t_step: time step (must satisfy the CFL max(lambda) < 2)
    :param h:
    :param g: metric tensor
    :return: vector x (at t = t_final)

    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    import numpy as np

    # parameters
    nb_step = np.ceil(t_final / t_step)  # number of time step
    t_step = t_final / nb_step

    # tensors
    detg = utils.tensor_determinant(g)
    detg = np.sqrt(detg)
    ginv = utils.tensor_inverse(g)
    ginv = utils.tensor_scalar_product(detg, ginv)
    detg2 = detg[1:-1, 1:-1]

    # LOOP
    x = x0
    for i in range(nb_step):
        m = t_step / h / h
        x = np.sum(x, -(
            utils.tensor_scalar_product(m,
                                        np.divide(utils.tensor_scalar_product(h, utils.operateur(x, ginv, detg)), detg2, dtype=object))))

    return x


def heat_solver_tensor_3D_P1_grad_conj(f, g, t_final, h, t_step, CL_value, epsilon):
    """
    It solves the poisson's equation in 1D on the regular mesh (with mesh of size h)
    :param f: approximation of a funcion of L^(/Omega)
    :param g: tensor
    :param t_final:
    :param h:
    :param t_step:
    :param CL_value:
    :param epsilon:
    :return: u= solution of the poisson's equation
    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    import numpy as np

    # initialisation
    if h is None:
        h = 1
    if CL_value is None:
        CL_value = np.zeros(f.shape)
    if epsilon is None:
        epsilon = 0.1

    # rigidity matrix
    b_h = f[1:-1, 1:-1, 1:-1] * (h * h * h)

    b_h[:, :, 0] = b_h[:, :, 0] + (
        CL_value[1:-1, 1:-1, 0] * h)  # not sure about b_h third value is 0 -> I need to avoid the column (HOW??)
    b_h[:, 0, :] = b_h[:, 0, :] + (CL_value[1:-1, 0, 1:-1] * h)
    b_h[0, :, :] = b_h[0, :, :] + (CL_value[0, 1:-1, 1:-1] * h)

    print('##########computation b_H#############@ ')

    # inversion of the linear system
    U_h = utils.heat_finite_elt_3D_tensor2(b_h, t_final, t_step, h, g)

    u = CL_value
    u[1:-1, 1:-1, 1:-1] = U_h

    return u


def heat_solver_tensor_2D_P1_grad_conj(f, g, t_final, h, t_step, CL_value, epsilon):
    """
    It solves the poisson's equation in 1D on the regular mesh (with mesh of size h)
    :param f: approximation of a funcion of L^(/Omega)
    :param g: tensor
    :param t_final:
    :param h:
    :param t_step:
    :param CL_value:
    :param epsilon:
    :return: u= solution of the poisson's equation
    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    import numpy as np

    # intiialisation
    if h is None:
        h = 1
    if CL_value is None:
        CL_value = np.zeros(f.shape)
    if epsilon is None:
        epsilon = 1e-4

    # rigidity matrix
    b_h = utils.tensor_scalar_product((h * h), f[1:-1, 1:-1])
    b_h[:, 0] = b_h[:, 0] + utils.tensor_scalar_product(h, CL_value[1:-1, 0])
    b_h[0, :] = b_h[0, :] + utils.tensor_scalar_product(h, CL_value[0, 1:-1])

    # inversion of the linear system
    U_h = utils.heat_finite_elt_2D_tensor2(b_h, t_final, t_step, h, g)

    u = CL_value
    u[1:-1, 1:-1] = U_h

    return u


def obtain_g_fisher_tensor(dartel_input, FWHM):
    """
    heat regularization based on the Fisher metric
    :param dartel_input: dartel template in MNI space
    :param sigma_loc: 10
    :param h: voxel size 1,5
    :param FWHM: mm of smoothing, parameters choosing by the user. default_value = 4
    :return: g: fisher tensor

    """

    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    import math
    import numpy as np
    import os

    #
    # PARAMETERS

    sigma_loc = 10
    error_tol = 0.001  # error for the estimation of the largest eigenvalue
    alpha_time = 0.9  # time_step = alpha_time * (time_step_max)
    max_proba = 0.999  # proba must be > 0 & < 1
    min_proba = 0.001
    h = 1.5  # voxel size

    # PARSE INPUTS/INIT
    sigma = FWHM / (2 * math.sqrt(2 * math.log(2)))  # sigma of voxels
    beta = sigma ** 2 / 2

    # SCALE MAPS
    xxx = []

    atlas = utils.atlas_decomposition(dartel_input)

    for i in atlas:

        image = utils.rescaleImage(i, [min_proba, max_proba])

        xxx.append(image)

    atlas = xxx
    si = atlas[0].shape

    # CREATE TENSOR
    g_atlas = utils.create_fisher_tensor(atlas)
    g_atlas = utils.tensor_scalar_product(h * h, g_atlas)
    g_pos = utils.tensor_eye(atlas)
    g_pos = utils.tensor_scalar_product(1 / float(sigma_loc ** 2), g_pos)

    g = utils.tensor_sum(g_atlas, g_pos)

    print("computing mean distance ... ")

    eigenv = utils.tensor_eigenvalues(g)

    print('done')

    dist_av = []

    for i in range(g.shape[0]):
        dist_av.append(np.sqrt(abs(eigenv[i])))
    dist_av = np.mean(dist_av)

    print("average distance ", dist_av)

    g = utils.tensor_scalar_product((1 / dist_av) / dist_av, g)

    np.save(os.path.abspath('./output_fisher_tensor.npy'), g)

    return g, os.path.abspath('./output_fisher_tensor.npy')


def obtain_time_step_estimation(dartel_input, FWHM, g):
    """

    :param h: 1,5 voxel size
    :param FWHM: mm of smoothing, defined by the user, default value = 4
    :param g: fisher tensor
    :return:
    """
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils

    import math
    import numpy as np
    import json
    import os
    import nibabel as nib

    # obtain voxel size with dartel_input
    head = nib.load(dartel_input)
    head_ = head.header
    for i in range(len(head_['pixdim'])):
        if head_['pixdim'][i] > 0:
            h = head_['pixdim'][i]

    error_tol = 0.001  # error for the estimation of the largest eigenvalue
    alpha_time = 0.9  # time_step = alpha_time * (time_step_max)
    sigma = FWHM / (2 * math.sqrt(2 * math.log(2)))  # sigma of voxels
    beta = sigma ** 2 / 2

    lam = utils.largest_eigenvalue_heat_3D_tensor2(g, h,
                                                   error_tol)
    print("lambda: ", lam)
    lam = np.array(lam.real, dtype='float64')
    t_step_max = 2 / lam

    t_step = alpha_time * t_step_max
    nbiter = np.ceil(beta / t_step)
    t_step = beta / nbiter

    # after t_step calculation: creation of json file
    data = {
        "MaxDeltaT": 0.0025,
        "Alpha": 0.9,
        "Epsilon": 1e-6,
        "BoundaryConditions": "TimeInvariant",
        "SigmaLoc": 10,
        "TimeStepMax": t_step_max,
        "SpatialPrior": "Tissues (GM, WM, CSF)",
        "RegularizationType": "Fisher",
        "FWHM": FWHM
    }

    json_data = json.dumps(data)
    with open('./output_data.json', 'w') as f:
        f.write(json_data)

    return t_step, os.path.abspath('./output_data.json')


def heat_solver_equation(input_image, g, FWHM, t_step, dartel_input):
    import math
    import clinica.pipelines.machine_learning_spatial_svm.spatial_svm_utils as utils
    import nibabel as nib
    import numpy as np
    import os

    # obtain voxel size with dartel_input
    head = nib.load(dartel_input)
    head_ = head.header
    for i in range(len(head_['pixdim'])):
        if head_['pixdim'][i] > 0:
            h = head_['pixdim'][i]

    sigma = FWHM / (2 * math.sqrt(2 * math.log(2)))  # sigma of voxels
    beta = sigma ** 2 / 2

    input_image_read = nib.load(input_image)
    input_image_data = input_image_read.get_data()
    input_image_data = np.array(input_image_data, dtype='float32')

    u = utils.heat_solver_tensor_3D_P1_grad_conj(input_image_data, g, beta, h, t_step, CL_value=None, epsilon=None)

    img = utils.spm_write_vol(input_image, u)

    nib.save(img, './regularized_' + os.path.basename(input_image))

    return os.path.abspath('./regularized_' + os.path.basename(input_image))



# coding: utf8
import clinica.utils.filemanip
import clinica.utils.participant

__author__ = "Alexis Guyot"
__copyright__ = "Copyright 2016-2019, The Aramis Lab Team"
__credits__ = ["Alexis Guyot"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Alexis Guyot"
__email__ = "alexis.guyot@icm-institute.org"
__status__ = "Development"


def check_template_reconalled(
        caps_dir, unique_subject_list, persubject_session_list2):
    """Check if the template creation has been run for each subject.

    Args:
        caps_dir (string): CAPS subdirectory containing the subjects
            (absolute path)
        unique_subject_list (list of string): list of unique
            participant_id
        persubject_session_list2 (list of string): list of per-subject
            list of sessions

    Returns:
        N/A
    """
    import os
    import subprocess
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as template_utils

    unique_subject_number = len(unique_subject_list)

    # check recon_all template creation for each subject
    for subject_index in range(unique_subject_number):
        subject = unique_subject_list[subject_index]
        session_list = persubject_session_list2[subject_index]
        # get template location in the CAPS directory
        caps_template_target = template_utils.get_capstemplate_path(
            caps_dir, subject, session_list)
        # check if template subfolder exists
        if not os.path.isdir(caps_template_target):
            error_msg = 'Error: {0} does not exist.'.format(
                caps_template_target)
            raise IOError(error_msg)
        # Check that the recon-all template creation run was successful
        log_file = os.path.join(caps_template_target, 'scripts', 'recon-all.log')
        if os.path.isfile(log_file):
            last_line = subprocess.check_output(['tail', '-1', log_file])
            if b'finished without error' not in last_line:
                error_msg = 'No template found for subject {0}.'.format(
                    subject)
                error_msg = '{0} Re-run t1-freesurfer-template.'.format(
                    error_msg)
                raise IOError(error_msg)


def check_reconalled(caps_dir, subject_list, session_list):
    """Check cross-sectional and -base recon-all run previously

    Checks if t1-freesurfer-cross-sectional ('recon-all -all') and if
    t1-freesurfer-template ('recon-all -base') have been correctly run
    beforehand.
    Will crash if any of those have failed (i.e., data are missing in
    the CAPS folder)

    Args:
        caps_dir (string): CAPS directory to contain the output
        subject_list (list): a list containing all the participant_id
        session_list (list): a list containing all the session_id

    Returns:
        N/A
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as template_utils
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_longitudinal_correction_utils as utils

    # Check if t1-freesurfer-cross-sectional has been correctly run
    # beforehand. Will crash if it has not
    template_utils.check_xsectional_reconalled(
        caps_dir, subject_list, session_list)

    # get unique list of subjects and corresponding per-subject sessions
    [unique_subject_list,
     persubject_session_list2] = clinica.utils.participant.get_unique_subjects(
         subject_list, session_list)

    # Check if t1-freesurfer-template has been correctly run
    # beforehand. Will crash if it has not
    utils.check_template_reconalled(
        caps_dir, unique_subject_list, persubject_session_list2)


def check_processed(
        subject_list,
        session_list,
        caps_dir,
        template_unpcssd_sublist):
    """Check which subjects have (/not) been processed already

    Check the .tsv file (to force overwrite already-processed subjects)
    generated while running template pipeline to see which subjects have
    been processed already and which have not yet.
    Additionally, this checks two extra things:
    1. that the user has not provided a new .tsv file after executing
    the template pipeline for running the longitudinal-correction
    2. that the CAPS folder does not find subjects for which a
    longitudinal-correction was condcuted with no corresponding template
    in CAPS dir (logical inconsistency)

    Args:
        subject_list (list of string): list of participant_id
        session_list (list of string): list of session_id corresponding
            to a specific participant_id
        caps_dir (string): location of CAPS folder
        template_unpcssd_sublist (list of string): list of all the
            subjects for which no processing was found in the CAPS
            folder

    Returns:
        unpcssd_subject_list (list of strings): list of all the subjects
            for which no template/longitudinal-correction processing was
            found in the CAPS folder
        unpcssd_session_list (list of list of strings): list of all the
            sessions for which no template/longitudinal-correction
            processing was found in the CAPS folder
        unpcssd_capstarget_list (string): list, for all subjects for
            which no template/longitudinal-correction processing was
            found in the CAPS folder, of associated CAPS folder where
            the template will get stored
        all_sublist (list of strings): list of all the subjects
        all_seslist (list of strings): list of all sessions
        all_capstargetlist (string): list, for all subjects and
            corresponding sessions of associated CAPS folder where the
            template will get stored
        rundir (str): path to the directory where clinica is being
            launched. Used to create a .tsv file in running directory
            if the user wishes to re-run the pipeline on subjects that
            had already been processed.
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as template_utils

    # Check processed/unprocessed subjects for longitudinal-correction
    [
        longcor_pcssd_sublist,
        dummy,
        dummy,
        longcor_unpcssd_sublist,
        longcor_unpcssd_seslist,
        longcor_unpcssd_capstargetlist,
        all_sublist,
        all_seslist,
        all_capstargetlist] = template_utils.check_caps_processing(
            'longitudinal-correction',
            subject_list,
            session_list,
            caps_dir)

    # Raise error if some subjects longitudinal-correction processed but
    # not template processed
    for longcor_pccsd_sub in longcor_pcssd_sublist:
        if longcor_pccsd_sub in template_unpcssd_sublist:
            error_msg = 'Error: longitudinal processing found for'
            error_msg = '{0} subject {1} '.format(error_msg, longcor_pccsd_sub)
            error_msg = '{0} but no corresponding template found.'.format(error_msg)
            raise ValueError(error_msg)

    # deduce list of unprocessed subjects/sessions (i.e., all
    # longitudinal-correction unprocessed subjects/sessions)
    unpcssd_sublist = longcor_unpcssd_sublist
    unpcssd_seslist = longcor_unpcssd_seslist
    unpcssd_capstargetlist = longcor_unpcssd_capstargetlist

    return [
        unpcssd_sublist,
        unpcssd_seslist,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist,
        all_capstargetlist]


def get_warning_msg(
        pcssd_capstargetlist,
        caps_dir,
        tsv_path,
        working_directory,
        n_procs):
    """Generate a warning message if processing in CAPS folder already

    Will generate a warning message to indicate to the user that there
    already is processing in the CAPS folder. Will also show the command
    to run if they wish to overwrite the existing CAPS data

    Args:
        processing_type (string): Either 'template' or 'longitudinal
            correction'. Used to build a warning message specific to the
            type of data that is being looked for in the CAPS folder.
        pcssd_capstargetlist (list of strings): list of CAPS subfolders
            which contain template / longitudinal correction data that
            have been processed already
        caps_dir (string): location of CAPS folder
        tsv_path (string): path to .tsv file provided by user. None if
            no subjects already processed.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        working_directory (string): path to working directory
            provided by user. None if they did not provide any.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        n_procs (int): number of processors provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.

    Returns:
        warning_msg (string): warning message
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as template_utils

    # define command line to force overwriting data in CAPS folder
    cl_overwrite_caps = template_utils.get_cl_overwritecaps(
        caps_dir, tsv_path, working_directory, n_procs)

    # sanity check: warning message only if at least one subject
    # processed already
    if pcssd_capstargetlist:
        # define warning message
        if len(pcssd_capstargetlist) == 1:
            warning_msg = 'There is already a template directory in:'
        if len(pcssd_capstargetlist) > 1:
            warning_msg = 'There are already template directories in:'
        warning_msg = '{0}\n'.format(warning_msg)
        for pcssd_capstarget in pcssd_capstargetlist:
            warning_msg = '{0}- {1}\n'.format(warning_msg, pcssd_capstarget)
        if len(pcssd_capstargetlist) == 1:
            warning_msg = '{0} To overwrite this directory, run:\n'.format(
                warning_msg)
        if len(pcssd_capstargetlist) > 1:
            warning_msg = '{0} To overwrite these directories, run:\n'.format(
                warning_msg)
        # add command line to force overwriting data in CAPS folder to
        # warning message
        warning_msg = '{0} \'{1}\''.format(warning_msg, cl_overwrite_caps)
    else:
        warning_msg = None

    return warning_msg


def to_process(
        pcssd_capstargetlist,
        unpcssd_sublist,
        unpcssd_seslist,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist,
        all_capstargetlist,
        in_caps_dir,
        in_overwrite_caps,
        in_working_directory,
        in_n_procs,
        overwrite_tsv_path):
    """ Get list of subjects to be processed

    The list of subjects to be processed will depend upon whether the
    user provided the option to --force-overwrite already existing
    processing or not. If the option --force-overwrite was not passed
    but existing processing was found, a warning message will be
    generated to show the command to run in order to force-overwrite the
    existing processing.

    Args:
        pcssd_capstargetlist (string): list, for all subjects for
            which template/longitudinal-correction processing was
            found in the CAPS folder, of associated CAPS folder where
            the template will get stored
        unpcssd_sublist (list of strings): list of all the subjects
            for which no processing was found in the CAPS folder
        unpcssd_seslist (list of objects): list, for all subjects
            for which no processing was found in the CAPS folder, of the
            list of associated session
        unpcssd_capstargetlist (string): list, for all subjects for
            which no processing was found in the CAPS folder, of
            associated CAPS folder where the template or
            longitudinal-correction processing will get stored
        all_sublist (list of strings): list of all the subjects
        all_seslist (list of objects): list, for all subjects of
            the list of associated sessions
        all_capstargetlist (string): list, for all subjects and
            corresponding sessions of associated CAPS folder where the
            template will get stored
        in_caps_dir (string): CAPS directory to contain the output
        in_overwrite_caps (string): Option provided by the user to state
            whether the data in the CAPS folder (i.e.,
            previously-computed template) should be overwritten or not
            by a new run. By default, do not overwrite. Will overwrite
            for the following values: 'true', 'True' and 'TRUE'
        in_working_directory (string): path to working directory
            provided by user. None if they did not provide any.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_n_procs (int): number of processors provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        overwrite_tsv_path (str): path to the .tsv file to use if the user
            wishes to re-run the pipeline on subjects that had already
            been processed. None if no subjects already processed.

    Returns:
        topcss_sublist (list of strings): list of the subjects that will
            be processed. All if --force-overwrite, else only subjects
            that have not been processed already
        topcss_seslist2 (list of list): list of the list of sessions
            that correspond to each subject to be process
        topcss_capstargetlist (list of strings): list, for all subjects
            to be processed, of the associated CAPS folder where the
            template will get stored
        overwrite_warning (string): warning message to display in case
            template) processing has been found in the CAPS folder for
            any of the input subject/session
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as template_utils
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_longitudinal_correction_utils as utils

    # Convert overwrite-caps flag to boolean
    force_overwrite = template_utils.boolean_overwrite_caps(in_overwrite_caps)

    # Generate warning message if some subjects have already been
    # processed and --force-overwrite option not provided
    overwrite_warning = None
    some_processed = len(unpcssd_sublist) != len(all_sublist)
    if (some_processed) and (not force_overwrite):
        overwrite_warning = utils.get_warning_msg(
            pcssd_capstargetlist,
            in_caps_dir,
            overwrite_tsv_path,
            in_working_directory,
            in_n_procs)

    # check what subjects / sessions / CAPS target dirs to process
    # depending on --force-overwrite flag
    if force_overwrite:
        topcss_sublist = all_sublist
        topcss_seslist2 = all_seslist
        topcss_capstargetlist = all_capstargetlist
    else:
        topcss_sublist = unpcssd_sublist
        topcss_seslist2 = unpcssd_seslist
        topcss_capstargetlist = unpcssd_capstargetlist

    return [
        topcss_sublist,
        topcss_seslist2,
        topcss_capstargetlist,
        overwrite_warning]


def process_input_node(
        in_caps_dir,
        in_subject_list,
        in_session_list,
        in_unpcssd_sublist,
        in_pcssd_capstargetlist,
        in_overwrite_caps,
        in_working_directory,
        in_n_procs,
        in_overwrite_tsv):
    """ Carry out all processing for the input node

    Performs all the processing for the input node:
    1. Check if t1-freesurfer-cross-sectional has been correctly run
        beforehand
    2. Check if there are already template subfolders in the CAPS dir.

    Args:
        in_caps_dir (string): CAPS directory to contain the output
        in_subject_list (list): a list containing all the participant_id
        in_session_list (list): a list containing all the session_id
        in_unpcssd_sublist (list of string): list of participant_id, where
            each participant appears only once, wich have not already
            been (template) processed
        in_pcssd_capstargetlist (list of string): list of path to the
            template directory in CAPS folder for each participant, for
            the participants that have already been processed
        in_overwrite_caps (string): Option provided by the user to state
            whether the data in the CAPS folder (i.e.,
            previously-computed template) should be overwritten or not
            by a new run. By default, do not overwrite. Will overwrite
            for the following values: 'true', 'True' and 'TRUE'
        in_working_directory (string): path to working directory
            provided by user. None if they did not provide any.
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_n_procs (int): number of processors provided by user
            Used to display the command line to force re-running the
            pipeline if a previous run was detected in CAPS folder.
        in_overwrite_tsv (str): path to the .tsv file to use if the user
            wishes to re-run the pipeline on subjects that had already
            been processed. None if no subjects already processed.

    Returns:
        out_subject_list (list of string): list of participant_id
        out_session_list (list of list): list of session_id associated
            to any participant
        out_capstarget_list (list of string): list of path to the
            template directory in CAPS folder for each participant
        out_overwrite_warning (string): warning message to display in
            case (template or longitudinal-correction) processing has
            been found in the CAPS folder for any of the input
            subject/session
        out_overwrite_sublist (list of strings): list of subjects for
            which either template or longitudinal processing should be
            re-done if the user re-runs the pipeline using the flag
            '--force_overwrite True'
        out_overwrite_seslist (list of strings): list of sessions for
            which either template or longitudinal processing should be
            re-done if the user re-runs the pipeline using the flag
            '--force_overwrite True'
    """
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_longitudinal_correction_utils as utils

    # check if we have the necessary recon-all data (cross-sectional
    # runs and created templates) prior to running longitudinal
    # processing. Will crash if we do not have this.
    utils.check_reconalled(in_caps_dir, in_subject_list, in_session_list)

    # Check target CAPS target dir for template /
    # longitudinal-correction processing before conducting any
    # longitudinal processing
    [
        unpcssd_sublist,
        unpcssd_seslist,
        unpcssd_capstargetlist,
        all_sublist,
        all_seslist,
        all_capstargetlist] = utils.check_processed(
            in_subject_list,
            in_session_list,
            in_caps_dir,
            in_unpcssd_sublist)

    # Deduce subjects to be processed next (depending on whether
    # '--force-overwrite' flag was provided or not)
    [
        out_subject_list,
        out_session_list,
        out_capstarget_list,
        out_overwrite_warning] = utils.to_process(
            in_pcssd_capstargetlist,
            unpcssd_sublist,
            unpcssd_seslist,
            unpcssd_capstargetlist,
            all_sublist,
            all_seslist,
            all_capstargetlist,
            in_caps_dir,
            in_overwrite_caps,
            in_working_directory,
            in_n_procs,
            in_overwrite_tsv)

    return [
        out_subject_list,
        out_session_list,
        out_capstarget_list,
        out_overwrite_warning]


def create_symlinks(
        in_caps_dir,
        in_subject_list,
        in_session_list,
        in_longsubdirnames_dict):
    """Create symlinks for recon-all execution

    t1-freesurfer-cross-sectional has created the following folders:
    [caps_dir]/subjects/sub-[sub_i]/ses-[ses_ij]/t1/...
    ...freesurfer_cross_sectional/sub-[sub_i]_ses-[ses_ij]
    and t1-freesurfer-template has created:
    [caps_dir]/subjects/sub-[sub_i]/long-[long_i]/...
    ...freesurfer_unbiased_template/sub-[sub_i]
    where i in range [1, subject_number], [long_i] corresponds to all
    seessions associated with [sub_i] and ij represents any of the
    potential sessions for subject i
    This function creates for all above folders the following symbolic
    link ./subjects/sub-[sub_i]_ses-[ses_ij]
    and
    ./subjects/sub-[sub_i] (template) in the working directory.
    All the symlinks are located in the same folder, disrespective of
    the subject (i.e. use a Node and not a MapNode). The path of that
    folder is returned so recon-all -base can later be run on all the
    sessions corresponding to subject i.

    Args:
        in_caps_dir (string): CAPS directory to contain the output
        in_subject_list (list of string): list of participant_id
        in_session_list (list of string): list of corresponding
            session_id
        in_longsubdirnames_dict (dictionary of strings): for each
            unique subject, name of longitudinal subofolder associated
            to the current set of visits

    Returns:
        out_subject_symlink_path (string): path to the folder
            containing all session symlinks and all template symlinks
            for all subject_id
    """
    import os
    import errno

    # change the relative path to be absolute path
    caps_path = os.path.expanduser(in_caps_dir)
    caps_dir = os.path.join(caps_path, 'subjects')

    # create subject-specific path in clinica's working directory
    # (later used as $SUBJECT_DIR)
    out_subject_symlink_path = os.path.abspath('./subjects')
    try:
        os.mkdir(out_subject_symlink_path)
    except OSError as oserror:
        if oserror.errno != errno.EEXIST:
            raise

    subject_number = len(in_subject_list)
    for subject_index in range(subject_number):
        # retrieve current {subject,session}
        subject = in_subject_list[subject_index]
        session = in_session_list[subject_index]

        # find all the sessions associated to the current subject
        long_subdir_name = in_longsubdirnames_dict[subject]

        # generate symlink for cross-sectional reconall run
        caps_subjectid_path = "{0}/{1}/{2}/t1/freesurfer_cross_sectional/{1}_{2}".format(
            caps_dir, subject, session)
        wd_subjectid_symlink_path = "{0}/{1}_{2}".format(
            out_subject_symlink_path, subject, session)
        if not os.path.exists(wd_subjectid_symlink_path):
            os.symlink(caps_subjectid_path, wd_subjectid_symlink_path)

        # generate symlink for the template associated with subject_id
        caps_template_path = "{0}/{1}/{2}/freesurfer_unbiased_template/{1}".format(
            caps_dir, subject, long_subdir_name)
        wd_template_symlink_path = "{0}/{1}".format(
            out_subject_symlink_path, subject)
        if not os.path.exists(wd_template_symlink_path):
            os.symlink(caps_template_path, wd_template_symlink_path)

    return out_subject_symlink_path


def get_reconalllong_flags(in_subject, in_session):
    """Get recon-all -long flags for each subject

    Creates the flags that are required to run the recon-all -long for
    each subject (i.e. run freesurfer longitudinal for each
    {subject;session}).

    Args:
        in_subject (string): participant_id (/!\ from self.subjects /!\)
        in_session (string): session_id (/!\ from self.sessions /!\)

    Returns:
        out_reconalllong_flags (string): all the flags to run
            recon-all -long for any particular subject and their
            (unique) associated session. There are as many
            'recon-all flags' as there are {subject, session}
    """
    template_id = in_subject
    out_reconalllong_flags = '-long {0}_{1} {2} -all'.format(
        in_subject, in_session, template_id)

    return out_reconalllong_flags


def check_reconall_longitudinal_single(subjects_dir,
                                       in_subject,
                                       in_session):
    """Check if longitudinal run successfully for any single subject

    Args:
        subjects_dir (string): CAPS subdirectory containing the subject
            (absolute path)
        in_subject (string): participant_id
        in_session (string): session_id

    Returns:
        out_longitudinal_result (Boolean): True if success, False
            otherwise
    """
    import os
    import subprocess

    # check if longitudinal subfolder exists
    subjectid_longitudinal = "{0}/{1}_{2}.long.{1}".format(
        subjects_dir, in_subject, in_session)
    if not os.path.isdir(subjectid_longitudinal):
        raise OSError(
            '{0} is not a valid folder.'.format(subjectid_longitudinal))

    # check the recon-all.log
    log_file = os.path.join(subjectid_longitudinal, 'scripts', 'recon-all.log')
    if os.path.isfile(log_file):
        last_line = subprocess.check_output(['tail', '-1', log_file])
        if b'finished without error' not in last_line:
            out_longitudinal_result = False
            ioerror_msg = 'Subject {0} has not been processed yet'.format(
                in_subject)
            ioerror_msg = '{0} Re-run t1-freesurfer-longitudinal.'.format(
                ioerror_msg)
            raise IOError(ioerror_msg)
        else:
            out_longitudinal_result = True
    else:
        raise IOError(
            'no recon-all.log script for subject {0}'.format(in_subject))

    return out_longitudinal_result


def run_reconalllong(in_subject,
                     in_session,
                     in_reconalllong_flags,
                     in_symlink_path):
    """Run recon-all -long for each subject

    Run freesurfer longitudinal for each {subject;session}). The
    recon-all command is built with previously computed arguments
    (flags).

    Args:
        in_subject (string): participant_id
        in_session (string): session_id
        in_reconalllong_flags (string): all the flags to run
            recon-all -long for any particular subject and their
            (unique) associated session
        in_symlink_path (string): paths to subject-specific folder
            containing all session symlinks for subject_id associated to
            in_reconalllong_flags.

    Returns:
        out_longitudinal_result (Boolean): True if success, False
            otherwise
    """
    import subprocess
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_longitudinal_correction_utils as utils

    # run recon-all -long
    reconalllong_command = 'recon-all {0} -sd {1}'.format(
        in_reconalllong_flags, in_symlink_path)
    subprocess_run_reconalllong = subprocess.run(
        reconalllong_command,
        shell=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    if subprocess_run_reconalllong.returncode != 0:
        raise ValueError('recon-all -long failed, returned non-zero code')

    # check whether recon-all -long failed or not
    out_longitudinal_result = utils.check_reconall_longitudinal_single(
        in_symlink_path, in_subject, in_session)

    return out_longitudinal_result


def write_stats_files(
        in_subject, in_session, in_symlink_path, in_longitudinal_result):
    """Generate statistics TSV files for a given {subject,session}.

    Statistics are derived from recon-all's own .stats files and are
    meant to provide a summary of those.

    Args:
        in_subject (string): participant_id
        in_session (string): session_id
        in_symlink_path (string): paths to subject-specific folder
            containing all session symlinks for subject_id associated to
            in_reconalllong_flags. Used for 1) retrieving the path to
            recon-all generated .tsv files 2) write stats files in a
            subfolder of the folder containing the symlink to
            {subject,session}
        in_longitudinal_result (Boolean): Used to force longitudinal
            recon-all to be run before generating the .tsv files from
            the longitudinal processing

    Returns:
        out_stats_path (string): path to the folder containing
            all the .tsv stats files for {subject,session}
    """
    import os
    from clinica.utils.freesurfer import generate_regional_measures

    # get path to longitudinal folder in working dir (where recon-all
    # results are stored, including .stats files)
    seg_path = os.path.expanduser(in_symlink_path)

    # define output dir for the .tsv files
    subses_long_id = '{0}_{1}.long.{0}'.format(in_subject, in_session)
    subses_long_stats_subdir = "{0}_stats".format(subses_long_id)
    out_stats_path = os.path.join(seg_path, subses_long_stats_subdir)

    # generate the .tsv files from .stats files
    subses_id = '{0}_{1}'.format(in_subject, in_session)
    generate_regional_measures(
        seg_path, subses_id, output_dir=out_stats_path, longitudinal=True)

    return out_stats_path


def copy_to_caps(
        in_subject,
        in_session,
        in_subject_dir,
        in_caps_target,
        in_caps_dir,
        in_overwrite_warning,
        in_overwrite_caps,
        in_stats_path):
    """ Copy template folder to CAPS dir

    Move the longitudinal-correction folder created by recon-all
    longitudinal to its correct location in the CAPS tree structure.

    Args:
        in_subject (string): participant_id
        in_session (string): session_id
        in_subject_dir (string): location of the subjects (in the
            working directory)
        in_caps_target (string): path to CAPS target subfolder for the
            longitudinal correction connected to the current
            {subject,session}
        in_caps_dir (string): CAPS directory to contain the output
            Used to make sure the folder we are overwriting (if
            overwrite-caps option provided) are subolders of the input
            CAPS dir
        in_overwrite_warning (string): warning message if some subjects
            have already been processed. Shows the subjects/sessions
            that have been processed and the command line to run in case
            the user wishes to re-run clinica on the subjects that have
            been processed already
        in_overwrite_caps (string): Option provided by the user to state
            whether the data in the CAPS folder (i.e.,
            previously-computed template) should be overwritten or not
            by a new run. By default, do not overwrite. Will overwrite
            for the following values: 'true', 'True' and 'TRUE'
        in_stats_path (string): path to the folder (in working
            directory) that contains

    Returns:
        out_copy2_caps (boolean): flag to indicate whether or not the
            copy was successful
    """
    import os
    import shutil
    import warnings
    import clinica.pipelines.t1_freesurfer_longitudinal.t1_freesurfer_template_utils as template_utils

    # Convert overwrite-caps flag to boolean
    force_overwrite = template_utils.boolean_overwrite_caps(in_overwrite_caps)

    # define longitudinal-correction location in the working directory
    wd_longitudinal_location = os.path.join(
        in_subject_dir, '{0}_{1}.long.{0}'.format(in_subject, in_session))

    # define location of .tsv stats files in the CAPS directory,
    # following a similar pattern to that of in_caps_target:
    # the CAPS target is as follows:
    # [caps_dir]/subjects/[subject]/[session]/t1/[longsubdir]/...
    # ...freesurfer_longitudinal/...
    # ...sub-[subject]_ses-[session].long.sub-[subject]
    # The CAPS stats dir differs as it is:
    # [caps_dir]/subjects/[subject]/[session]/t1/[longsubdir]/...
    # ...freesurfer_longitudinal/regional_measures
    caps_target_stats = os.path.join(
        os.path.dirname(in_caps_target),
        'regional_measures')

    # Check if overwrite-caps option provided. If provided, delete the
    # target subdirectory in CAPS folder
    if force_overwrite:
        # check the target subdirectory exists in CAPS
        if os.path.exists(in_caps_target):
            # remove
            template_utils.safer_rmtree(in_caps_target, in_caps_dir)
        # check the stats subdirectory exists in CAPS
        if os.path.exists(caps_target_stats):
            # remove
            template_utils.safer_rmtree(caps_target_stats, in_caps_dir)

    # copy the longitudinal processing from working directory to CAPS
    # directory
    shutil.copytree(wd_longitudinal_location, in_caps_target)

    # copy the .tsv stats files from working directory to CAPS directory
    shutil.copytree(in_stats_path, caps_target_stats)
    out_copy2_caps = True

    # show warning message
    if in_overwrite_warning is not None:
        warnings.warn(in_overwrite_warning)

    return out_copy2_caps



# coding: utf8


def read_psf_information(psf_tsv, subject_ids, session_ids):
    import os
    from pandas.io.parsers import read_csv

    if not os.path.isfile(psf_tsv):
        raise FileNotFoundError('Could not find the psf_tsv file %s' % psf_tsv)
    try:
        psf_df = read_csv(psf_tsv, sep='\t')
    except (IOError, UnicodeDecodeError):
        raise RuntimeError('An error while reading %s happened' % psf_tsv)

    if psf_df.shape[0] != len(subject_ids):
        raise ValueError('The number of rows in fwhm_tsv file must match the number of subject-session pairs.')

    if any(elem not in ['participant_id', 'session_id', 'fwhm_x', 'fwhm_y', 'fwhm_z'] for elem in list(psf_df.columns)):
        raise IOError('The file %s must contain the following columns (separated by tabulations):\n'
                      'participant_id, session_id, fwhm_x, fwhm_y, fwhm_z\n'
                      'But we found:\n'
                      '%s\n'
                      'Pay attention to the spaces (there should be none).' %
                      (psf_tsv, str(list(psf_df.columns))))

    subjects_fwhm = list(psf_df.participant_id)
    sessions_fwhm = list(psf_df.session_id)
    idx_reordered = []
    for i, sub in enumerate(subject_ids):
        current_ses = session_ids[i]
        idx_sub = [j for j in range(len(subjects_fwhm)) if sub == subjects_fwhm[j] and current_ses == sessions_fwhm[j]]
        if len(idx_sub) == 0:
            raise RuntimeError('Subject %s with session %s that you want to proceed was not found '
                               'in the TSV file containing PSF specifications (%s).' %
                               (sub, current_ses, psf_tsv))
        if len(idx_sub) > 1:
            raise RuntimeError('Subject %s with session %s that you want to proceed was found multiple times '
                               'in the TSV file containing PSF specifications (%s).' %
                               (sub, current_ses, psf_tsv))
        idx_reordered.append(idx_sub[0])

    fwhm_x = list(psf_df.fwhm_x)
    fwhm_y = list(psf_df.fwhm_y)
    fwhm_z = list(psf_df.fwhm_z)
    iterables_fwhm = [[fwhm_x[i], fwhm_y[i], fwhm_z[i]] for i in idx_reordered]
    return iterables_fwhm


def create_binary_mask(tissues, threshold=0.3):
    """

    Args:
        tissues:
        threshold:

    Returns:

    """
    import nibabel as nib
    import numpy as np
    from os import getcwd
    from os.path import join, basename

    if len(tissues) == 0:
        raise RuntimeError('The length of the list of tissues must be greater than zero.')

    img_0 = nib.load(tissues[0])
    shape = list(img_0.get_data().shape)

    data = np.zeros(shape=shape)

    for image in tissues:
        data = data + nib.load(image).get_data()

    data = (data > threshold) * 1.0
    out_mask = join(getcwd(), basename(tissues[0]) + '_brainmask.nii')

    mask = nib.Nifti1Image(data, img_0.affine, header=img_0.header)
    nib.save(mask, out_mask)
    return out_mask


def apply_binary_mask(image, binary_mask):
    import nibabel as nib
    from os import getcwd
    from os.path import join, basename

    original_image = nib.load(image)
    mask = nib.load(binary_mask)

    data = original_image.get_data() * mask.get_data()

    masked_image_path = join(getcwd(), 'masked_' + basename(image))
    masked_image = nib.Nifti1Image(data, original_image.affine, header=original_image.header)
    nib.save(masked_image, masked_image_path)
    return masked_image_path


def create_pvc_mask(tissues):

    import nibabel as nib
    import numpy as np
    from os import getcwd
    from os.path import join

    if len(tissues) == 0:
        raise RuntimeError('The length of the list of tissues must be greater than zero.')

    img_0 = nib.load(tissues[0])
    shape = img_0.get_data().shape
    background = np.zeros(shape=shape)

    shape += tuple([len(tissues) + 1])
    data = np.empty(shape=shape, dtype=np.float64)

    for i in range(len(tissues)):
        image = nib.load(tissues[i])
        data[..., i] = np.array(image.get_data())
        background = background + image.get_data()

    background = 1.0 - background
    data[..., len(tissues)] = np.array(background)

    out_mask = join(getcwd(), 'pvc_mask.nii')
    mask = nib.Nifti1Image(data, img_0.affine, header=img_0.header)
    nib.save(mask, out_mask)
    return out_mask


def pet_pvc_name(pet_image, pvc_method):
    from os.path import basename
    pet_pvc_path = 'pvc-' + pvc_method.lower() + '_' + basename(pet_image)
    return pet_pvc_path


def normalize_to_reference(pet_image, region_mask):
    import nibabel as nib
    import numpy as np
    from os import getcwd
    from os.path import basename, join

    pet = nib.load(pet_image)
    ref = nib.load(region_mask)

    region = np.multiply(pet.get_data(), ref.get_data())
    region_mean = np.nanmean(np.where(region != 0, region, np.nan))

    data = pet.get_data() / region_mean

    suvr_pet_path = join(getcwd(), 'suvr_' + basename(pet_image))

    suvr_pet = nib.Nifti1Image(data, pet.affine, header=pet.header)
    nib.save(suvr_pet, suvr_pet_path)

    return suvr_pet_path


def atlas_statistics(in_image, in_atlas_list):
    """
    For each atlas name provided it calculates for the input image the mean
    for each region in the atlas and saves it to a TSV file.

    Args:
        in_image: A Nifti image
        in_atlas_list: List of names of atlas to be applied

    Returns:
        List of paths to tsv files
    """
    from os import getcwd
    from os.path import abspath, join
    from nipype.utils.filemanip import split_filename
    from clinica.utils.atlas import AtlasAbstract
    from clinica.utils.statistics import statistics_on_atlas

    orig_dir, base, ext = split_filename(in_image)
    atlas_classes = AtlasAbstract.__subclasses__()
    atlas_statistics_list = []
    for atlas in in_atlas_list:
        for atlas_class in atlas_classes:
            if atlas_class.get_name_atlas() == atlas:
                out_atlas_statistics = abspath(join(getcwd(), base + '_space-' + atlas + '_statistics.tsv'))
                statistics_on_atlas(in_image, atlas_class(), out_atlas_statistics)
                atlas_statistics_list.append(out_atlas_statistics)
                break

    return atlas_statistics_list


def pet_container_from_filename(pet_filename):
    import re
    from os.path import join
    m = re.search(r'(sub-[a-zA-Z0-9]+)_(ses-[a-zA-Z0-9]+)_', pet_filename)

    if m is None:
        raise ValueError('Input filename is not in a BIDS or CAPS compliant format. It does not contain the subject' +
                         ' and session information.')

    subject = m.group(1)
    session = m.group(2)

    return join('subjects', subject, session, 'pet/preprocessing')


def get_from_list(in_list, index):
    return in_list[index]



# coding: utf8

import clinica.engine as ce

__author__ = "Alexis Guyot"
__copyright__ = "Copyright 2016-2019, The Aramis Lab Team"
__credits__ = ["Alexis Guyot"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Alexis Guyot"
__email__ = "alexis.guyot@icm-institute.org"
__status__ = "Development"


class T1FreeSurferLongitudinalCLI(ce.CmdParser):

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 't1-freesurfer-longitudinal'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = ('Longitudinal pre-processing of T1w images with FreeSurfer:\n'
                             'http://clinica.run/doc/Pipelines/T1_FreeSurfer_Longitudinal/')

    def define_options(self):
        """Define the sub-command arguments
        """
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES
        # clinica compulsory arguments (e.g. bids, caps, group_id)
        clinica_comp = self._args.add_argument_group(PIPELINE_CATEGORIES['CLINICA_COMPULSORY'])
        clinica_comp.add_argument("caps_directory",
                                  help='Path to the CAPS directory.')
        # clinica standard arguments (e.g. --n_procs)
        clinica_opt = self.add_clinica_standard_arguments()
        # clinica pipeline-specific arguments
        clinica_opt.add_argument("-oc", "--overwrite_caps",
                                 type=str, default='false',
                                 help='Overwrite existing data in CAPS directory '
                                      '(Possible values: true, True, TRUE, false, False, FALSE. Default: -oc false)')

    def run_command(self, args):
        """Run the pipelines with defined args
        """
        import nipype.pipeline.engine as npe
        from nipype.pipeline.engine import Workflow
        from .t1_freesurfer_template_pipeline import T1FreeSurferTemplate
        from .t1_freesurfer_longitudinal_correction_pipeline import T1FreeSurferLongitudinalCorrection

        template_pipeline = T1FreeSurferTemplate(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory)
        )
        longcorr_pipeline = T1FreeSurferLongitudinalCorrection(
            caps_directory=self.absolute_path(args.caps_directory),
            tsv_file=self.absolute_path(args.subjects_sessions_tsv),
            base_dir=self.absolute_path(args.working_directory)
        )

        # add n_procs to pipeline parameters so we
        # know, while running the pipeline, which arguments the user
        # passed to the CLI
        template_pipeline.parameters = {
            # Todo: Remove parameters['n_procs']
            'n_procs': args.n_procs,
            'overwrite_caps': args.overwrite_caps
            }
        longcorr_pipeline.parameters = {
            # Todo: Remove parameters['n_procs']
            'n_procs': args.n_procs,
            'overwrite_caps': args.overwrite_caps
            }

        # separately build the two template and longitudinal-correction pipelines
        template_pipeline.build()
        longcorr_pipeline.build()

        # create overall workflow
        longitudinal_workflow = npe.Workflow(name='T1FreeSurferLongitudinal')
        longitudinal_workflow.base_dir = self.absolute_path(args.working_directory)
        # connect the template pipeline to the longitudinal-correction pipeline
        longitudinal_workflow.connect(
            template_pipeline, '5_sendto_longcorr.out_unpcssd_sublist',
            longcorr_pipeline, '0_receivefrom_template.unpcssd_sublist')
        longitudinal_workflow.connect(
            template_pipeline, '5_sendto_longcorr.out_pcssd_capstargetlist',
            longcorr_pipeline, '0_receivefrom_template.pcssd_capstargetlist')
        longitudinal_workflow.connect(
            template_pipeline, '5_sendto_longcorr.out_overwrite_tsv',
            longcorr_pipeline, '0_receivefrom_template.overwrite_tsv')
        # run overall workflow
        if args.n_procs:
            Workflow.run(
                longitudinal_workflow,
                plugin_args={'n_procs': args.n_procs},
                plugin='MultiProc')
        else:
            Workflow.run(longitudinal_workflow)



# coding: utf-8

"""
 Module for converting Tau PET of ADNI
"""

__author__ = "Jorge Samper-Gonzalez"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__credits__ = ["Sabrina Fontanella"]
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_tau_pet(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert Tau PET images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from clinica.iotools.converters.adni_to_bids.adni_utils import t1_pet_paths_to_bids
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of TAU PET images. Output will be stored in %s.' % path.join(dest_dir, 'conversion_info'))
    images = compute_tau_pet_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of TAU PET images found. Exporting images into BIDS ...')
    t1_pet_paths_to_bids(images, dest_dir, 'tau')
    cprint(Fore.GREEN + 'TAU PET conversion done.' + Fore.RESET)


def compute_tau_pet_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute the paths to Tau PET images

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns: pandas Dataframe containing the path for each Tau PET image

    """

    import pandas as pd
    import os
    import operator
    from os import path
    from functools import reduce
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, find_image_path
    from clinica.utils.stream import cprint

    pet_tau_col = ['Phase', 'Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date', 'Study_ID',
                   'Series_ID', 'Image_ID', 'Original']
    pet_tau_df = pd.DataFrame(columns=pet_tau_col)

    # Loading needed .csv files
    tauqc = pd.read_csv(path.join(csv_dir, 'TAUQC.csv'), sep=',', low_memory=False)
    tauqc3 = pd.read_csv(path.join(csv_dir, 'TAUQC3.csv'), sep=',', low_memory=False)
    pet_meta_list = pd.read_csv(path.join(csv_dir, 'PET_META_LIST.csv'), sep=',', low_memory=False)

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list['Subject'] == subj]

        if subject_pet_meta.shape[0] < 1:
            # TODO Log somewhere subjects without TAU PET images metadata
            continue

        # QC for TAU PET images for ADNI 2
        tau_qc2_subj = tauqc[(tauqc.SCANQLTY == 1) & (tauqc.RID == int(subj[-4:]))]

        # QC for TAU PET images for ADNI 3
        tau_qc3_subj = tauqc3[(tauqc3.SCANQLTY == 1) & (tauqc3.RID == int(subj[-4:]))]

        # Concatenating visits in both QC files
        tau_qc_subj = pd.concat([tau_qc2_subj, tau_qc3_subj], axis=0, ignore_index=True, sort=False)

        for visit in list(tau_qc_subj.VISCODE2.unique()):
            # TODO Infer visit from ADNIMERGE visits
            if str(visit) == 'nan':
                continue

            pet_qc_visit = tau_qc_subj[tau_qc_subj.VISCODE2 == visit]

            # If there are several scans for a timepoint we keep image acquired last (higher LONIUID)
            pet_qc_visit = pet_qc_visit.sort_values("LONIUID", ascending=False)

            qc_visit = pet_qc_visit.iloc[0]

            # Corresponding LONI image ID for original scan in PET Meta List
            int_image_id = int(qc_visit.LONIUID[1:])

            original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                 & (subject_pet_meta['Image ID'] == int_image_id)]

            # If no corresponding TAU PET metadata for an original image,
            # take scan at the same date containing TAU or AV 45 in sequence name
            if original_pet_meta.shape[0] < 1:
                original_pet_meta = subject_pet_meta[(subject_pet_meta['Orig/Proc'] == 'Original')
                                                     & subject_pet_meta.Sequence.map(
                    lambda x: ((x.lower().find('tau') > -1) |
                               (x.lower().find('av-1451') > -1) |
                               (x.lower().find('av1451') > -1)))
                                                        & (subject_pet_meta['Scan Date'] == qc_visit.SCANDATE)]

                if original_pet_meta.shape[0] < 1:
                    # TODO Log somewhere QC visits without image metadata
                    cprint('No TAU-PET images metadata for subject - ' + subj + ' for visit ' + qc_visit.VISCODE2)
                    continue

            original_image = original_pet_meta.iloc[0]

            # Co-registered and Averaged image with the same Series ID of the original image
            averaged_pet_meta = subject_pet_meta[(subject_pet_meta['Sequence'] == 'AV1451 Co-registered, Averaged')
                                                 & (subject_pet_meta['Series ID'] == original_image['Series ID'])]

            # If an explicit AV1451 Co-registered, Averaged image does not exist,
            # the original image is already in that preprocessing stage.

            if averaged_pet_meta.shape[0] < 1:
                sel_image = original_image
                original = True
            else:
                sel_image = averaged_pet_meta.iloc[0]
                original = False

            visit = sel_image.Visit
            sequence = replace_sequence_chars(sel_image.Sequence)
            date = sel_image['Scan Date']
            study_id = sel_image['Study ID']
            series_id = sel_image['Series ID']
            image_id = sel_image['Image ID']

            row_to_append = pd.DataFrame(
                [[qc_visit.Phase, subj, qc_visit.VISCODE2, str(visit), sequence, date, str(study_id), str(series_id),
                  str(image_id), original]],
                columns=pet_tau_col)
            pet_tau_df = pet_tau_df.append(row_to_append, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Multiple output images
                         ('098_S_4275', 'm84')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((pet_tau_df.Subject_ID == conv_error[0])
                             & (pet_tau_df.VISCODE == conv_error[1]))

    if error_indices:
        indices_to_remove = pet_tau_df.index[reduce(operator.or_, error_indices, False)]
        pet_tau_df.drop(indices_to_remove, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(pet_tau_df, source_dir, 'TAU', 'I', 'Image_ID')

    tau_csv_path = path.join(dest_dir, 'conversion_info')
    if not os.path.exists(tau_csv_path):
        os.mkdir(tau_csv_path)
    images.to_csv(path.join(tau_csv_path, 'tau_pet_paths.tsv'), sep='\t', index=False)

    return images


def check_exceptions(bids_dir):
    from os import path
    import pandas as pd
    from glob import glob

    tau_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 'tau_pet_paths.tsv'), sep='\t')

    tau_paths = tau_paths[tau_paths.Path.notnull()]

    tau_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in tau_paths.Subject_ID.to_list()]
    tau_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in tau_paths.VISCODE.to_list()]

    count = 0

    for r in tau_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'pet')
        image_pattern = path.join(image_dir, '%s_%s_*tau*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)
        if not files_list:
            print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            count += 1

        elif len(files_list) > 1:
            print("Too many images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            print(files_list)

    print(count)



# coding: utf-8

"""
 Module for converting DWI of ADNI
"""

__author__ = "Jorge Samper-Gonzalez and Sabrina Fontanella"
__copyright__ = "Copyright 2016-2019 The Aramis Lab Team"
__license__ = "See LICENSE.txt file"
__version__ = "0.1.0"
__maintainer__ = "Jorge Samper-Gonzalez"
__email__ = "jorge.samper-gonzalez@inria.fr"
__status__ = "Development"


def convert_adni_dwi(source_dir, csv_dir, dest_dir, subjs_list=None):
    """
    Convert DW images of ADNI into BIDS format

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    """

    import pandas as pd
    from os import path
    from clinica.utils.stream import cprint
    from colorama import Fore

    if subjs_list is None:
        adni_merge_path = path.join(csv_dir, 'ADNIMERGE.csv')
        adni_merge = pd.read_csv(adni_merge_path, sep=',', low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint('Calculating paths of DWI images. Output will be stored in ' + path.join(dest_dir, 'conversion_info') + '.')
    images = compute_dwi_paths(source_dir, csv_dir, dest_dir, subjs_list)
    cprint('Paths of DWI images found. Exporting images into BIDS ...')
    dwi_paths_to_bids(images, dest_dir)
    cprint(Fore.GREEN + 'DWI conversion done.' + Fore.RESET)


def compute_dwi_paths(source_dir, csv_dir, dest_dir, subjs_list):
    """
    Compute paths to DW images to convert to BIDS

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list

    Returns:
        images: pandas dataframe that contains the path to all the DW images to convert

    """

    import operator
    from os import path, mkdir
    from functools import reduce
    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import find_image_path, visits_to_timepoints_mrilist

    dwi_col_df = ['Subject_ID', 'VISCODE', 'Visit', 'Sequence', 'Scan_Date',
                  'Study_ID', 'Series_ID', 'Image_ID', 'Field_Strength']
    dwi_df = pd.DataFrame(columns=dwi_col_df)

    # Loading needed .csv files
    adni_merge = pd.read_csv(path.join(csv_dir, 'ADNIMERGE.csv'), sep=',', low_memory=False)

    mayo_mri_qc = pd.read_csv(path.join(csv_dir, 'MAYOADIRL_MRI_IMAGEQC_12_08_15.csv'), sep=',', low_memory=False)
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == 'DTI']

    mri_list = pd.read_csv(path.join(csv_dir, 'MRILIST.csv'), sep=',', low_memory=False)

    # Selecting only DTI images that are not Multiband, processed or enchanced images
    mri_list = mri_list[mri_list.SEQUENCE.map(lambda x: x.lower().find('dti') > -1)]
    unwanted_sequences = ['MB', 'ADC', 'FA', 'TRACEW', 'Enhanced', 'Reg']
    mri_list = mri_list[mri_list.SEQUENCE.map(lambda x: not any(subs in x for subs in unwanted_sequences))]

    for subj in subjs_list:

        # Filter ADNIMERGE, MRI_LIST and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values('EXAMDATE')

        mri_list_subj = mri_list[mri_list.SUBJECT == subj]
        mri_list_subj = mri_list_subj.sort_values('SCANDATE')

        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints_mrilist(subj, mri_list_subj, adnimerge_subj, "DWI")

        for visit_info in visits.keys():
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            visit_mri_list = mri_list_subj[mri_list_subj.VISIT == visit_str]
            axial = dwi_image(subj, timepoint, visits[visit_info], visit_mri_list, mayo_mri_qc_subj)

            if axial is not None:
                row_to_append = pd.DataFrame(axial, index=['i', ])
                # TODO Replace iteratively appending by pandas.concat
                dwi_df = dwi_df.append(row_to_append, ignore_index=True, sort=False)

    # Exceptions
    # ==========
    conversion_errors = [('029_S_2395', 'm60'),
                         ('029_S_0824', 'm108'),
                         ('029_S_0914', 'm108'),
                         ('027_S_2219', 'm36'),
                         ('129_S_2332', 'm12'),
                         ('029_S_4384', 'm48'),
                         ('029_S_4385', 'm48'),
                         ('029_S_4585', 'm48'),
                         ('016_S_4591', 'm24'),
                         ('094_S_4630', 'm06'),
                         ('094_S_4649', 'm06'),
                         ('029_S_5219', 'm24'),
                         ('094_S_2238', 'm48'),
                         ('129_S_4287', 'bl'),
                         ('007_S_4611', 'm03'),
                         ('016_S_4638', 'bl'),
                         ('027_S_5118', 'bl'),
                         ('098_S_4018', 'bl'),
                         ('098_S_4003', 'm12'),
                         ('016_S_4584', 'm24'),
                         ('016_S_5007', 'm12'),
                         ('129_S_2347', 'm06'),
                         ('129_S_4220', 'bl'),
                         ('007_S_2058', 'm12'),
                         ('016_S_2007', 'm06'),
                         ('020_S_6358', 'bl'),
                         ('114_S_6039', 'm12'),
                         ('114_S_6057', 'bl'),
                         ('153_S_6274', 'bl'),
                         ('006_S_4485', 'm84'),
                         ('153_S_6237', 'bl'),
                         ('153_S_6336', 'bl'),
                         ('153_S_6450', 'bl')]

    error_indices = []
    for conv_error in conversion_errors:
        error_indices.append((dwi_df.Subject_ID == conv_error[0])
                             & (dwi_df.VISCODE == conv_error[1]))

    if error_indices:
        indices_to_remove = dwi_df.index[reduce(operator.or_, error_indices, False)]
        dwi_df.drop(indices_to_remove, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(dwi_df, source_dir, 'DWI', 'S', 'Series_ID')

    dwi_tsv_path = path.join(dest_dir, 'conversion_info')
    if not path.exists(dwi_tsv_path):
        mkdir(dwi_tsv_path)
    images.to_csv(path.join(dwi_tsv_path, 'dwi_paths.tsv'), sep='\t', index=False)

    return images


def dwi_paths_to_bids(images, dest_dir, mod_to_update=False):
    """
    Convert DWI images

    Args:
        images: dataframe returned by the method compute_dwi_paths
        dest_dir: path to the destination directory
        mod_to_update: if is true and an image is already existing it will overwrite the old version

    """
    from multiprocessing import cpu_count, Pool, Value
    from functools import partial

    counter = None

    def init(args):
        ''' store the counter for later use '''
        global counter
        counter = args

    subjs_list = images['Subject_ID'].unique()
    counter = Value('i', 0)
    total = len(images)
    partial_generate_subject_files = partial(generate_subject_files,
                                             images=images,
                                             dest_dir=dest_dir,
                                             mod_to_update=mod_to_update)
    poolrunner = Pool(cpu_count(), initializer=init, initargs=(counter,))
    poolrunner.map(partial_generate_subject_files, subjs_list)
    del counter


def dwi_image(subject_id, timepoint, visit_str, visit_mri_list, mri_qc_subj):
    """

    Args:
        subject_id:
        timepoint:
        visit_str:
        visit_mri_list:
        mri_qc_subj:

    Returns:

    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import replace_sequence_chars, select_image_qc

    sel_image = select_image_qc(list(visit_mri_list.IMAGEUID), mri_qc_subj)
    if sel_image is None:
        return None

    sel_scan = visit_mri_list[visit_mri_list.IMAGEUID == sel_image].iloc[0]

    image_dict = {'Subject_ID': subject_id,
                  'VISCODE': timepoint,
                  'Visit': visit_str,
                  'Sequence': replace_sequence_chars(sel_scan.SEQUENCE),
                  'Scan_Date': sel_scan['SCANDATE'],
                  'Study_ID': str(int(sel_scan.STUDYID)),
                  'Series_ID': str(int(sel_scan.SERIESID)),
                  'Image_ID': str(int(sel_scan.IMAGEUID)),
                  'Field_Strength': sel_scan.MAGSTRENGTH}

    return image_dict


def generate_subject_files(subj, images, dest_dir, mod_to_update):
    import clinica.iotools.bids_utils as bids
    import clinica.iotools.converters.adni_to_bids.adni_utils as adni_utils
    from clinica.utils.stream import cprint
    import subprocess
    import os
    import shutil
    from os import path
    from glob import glob

    global counter
    alpha_id = bids.remove_space_and_symbols(subj)
    bids_id = 'sub-ADNI' + alpha_id
    # Extract the list of sessions available from the dwi paths files, removing the duplicates
    sess_list = images[(images['Subject_ID'] == subj)]['VISCODE'].unique()

    if not os.path.exists(path.join(dest_dir, bids_id)):
        os.mkdir(path.join(dest_dir, bids_id))

    # For each session available, create the folder if doesn't exist and convert the files
    for ses in sess_list:
        with counter.get_lock():
            counter.value += 1
        cprint('[DWI] Processing subject ' + str(subj)
               + ' - session ' + ses + ', ' + str(counter.value)
               + ' / ' + str(len(images)))
        ses_bids = adni_utils.viscode_to_session(ses)
        bids_ses_id = 'ses-' + ses_bids
        bids_file_name = bids_id + '_ses-' + ses_bids
        ses_path = path.join(dest_dir, bids_id, bids_ses_id)

        if mod_to_update:
            if os.path.exists(path.join(ses_path, 'dwi')):
                shutil.rmtree(path.join(ses_path, 'dwi'))

        if not os.path.exists(ses_path):
            os.mkdir(ses_path)

        dwi_info = images[
            (images['Subject_ID'] == subj) & (images['VISCODE'] == ses)]

        # For the same subject, same session there could be multiple dwi with different acq label
        for j in range(len(dwi_info)):
            dwi_subj = dwi_info.iloc[j]
            if type(dwi_subj['Path']) != float and dwi_subj['Path'] != '':
                if not os.path.exists(path.join(ses_path, 'dwi')):
                    os.mkdir(path.join(ses_path, 'dwi'))
                dwi_path = dwi_subj['Path']

                bids_name = bids_file_name + '_acq-axial_dwi'

                bids_dest_dir = path.join(ses_path, 'dwi')

                if not os.path.exists(bids_dest_dir):
                    os.mkdir(dest_dir)
                command = 'dcm2niix -b n -z y -o ' + bids_dest_dir + ' -f ' + bids_name + ' ' + dwi_path
                subprocess.run(command,
                               shell=True,
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)

                # Removing ADC images
                adc_image = path.join(bids_dest_dir, bids_name + '_ADC.nii.gz')
                if os.path.exists(adc_image):
                    os.remove(adc_image)

                # If dcm2niix didn't work use dcm2nii
                # print path.join(dest_dir, bids_name + '.nii.gz')
                if not os.path.exists(path.join(bids_dest_dir,
                                                bids_name + '.nii.gz')) or not os.path.exists(
                        path.join(bids_dest_dir,
                                  bids_name + '.bvec') or not os.path.exists(
                            path.join(bids_dest_dir,
                                      bids_name + '.bval'))):
                    cprint('\tConversion with dcm2niix failed, trying with dcm2nii')

                    # Find all the files eventually created by dcm2niix and remove them
                    dwi_dcm2niix = glob(path.join(bids_dest_dir, bids_name + '*'))

                    for d in dwi_dcm2niix:
                        # print 'Removing the old', d
                        os.remove(d)

                    command = 'dcm2nii -a n -d n -e n -i y -g y -p n -m n -r n -x n -o %s %s' \
                              % (bids_dest_dir, dwi_path)
                    subprocess.run(command,
                                   shell=True,
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.DEVNULL)
                    nii_file = path.join(bids_dest_dir,
                                         subj.replace('_', '') + '.nii.gz')
                    bvec_file = path.join(bids_dest_dir,
                                          subj.replace('_', '') + '.bvec')
                    bval_file = path.join(bids_dest_dir,
                                          subj.replace('_', '') + '.bval')

                    if os.path.exists(bvec_file) and os.path.exists(
                            bval_file):
                        os.rename(bvec_file, path.join(bids_dest_dir,
                                                       bids_name + '.bvec'))
                        os.rename(bval_file, path.join(bids_dest_dir,
                                                       bids_name + '.bval'))
                    else:
                        cprint('WARNING: bvec and bval not generated by dcm2nii'
                               + ' for subject ' + subj + ' and session ' + ses)

                    if os.path.exists(nii_file):
                        os.rename(nii_file, path.join(bids_dest_dir,
                                                      bids_name + '.nii.gz'))
                    else:
                        cprint('WARNING: CONVERSION FAILED...'
                               + ' for subject ' + subj + ' and session ' + ses)


def check_exceptions(bids_dir):
    from os import path
    import pandas as pd
    from glob import glob

    dwi_paths = pd.read_csv(path.join(bids_dir, 'conversion_info', 'dwi_paths.tsv'), sep='\t')

    dwi_paths = dwi_paths[dwi_paths.Path.notnull()]

    dwi_paths['BIDS_SubjID'] = ['sub-ADNI' + s.replace('_', '') for s in dwi_paths.Subject_ID.to_list()]
    dwi_paths['BIDS_Session'] = ['ses-' + s.replace('bl', 'm00').upper() for s in dwi_paths.VISCODE.to_list()]

    count = 0
    count_ADC = 0

    for r in dwi_paths.iterrows():
        image = r[1]
        image_dir = path.join(bids_dir, image.BIDS_SubjID, image.BIDS_Session, 'dwi')
        image_pattern = path.join(image_dir, '%s_%s_*' % (image.BIDS_SubjID, image.BIDS_Session))
        files_list = glob(image_pattern)
        if not files_list:
            print("No images for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
            # count += 1

        elif len(files_list) != 3:
            if sum(["ADC" in f for f in files_list]):
                count_ADC += 1
            else:
                print("Wrong files count for subject %s in session %s" % (image.BIDS_SubjID, image.BIDS_Session))
                print(files_list)
                count += 1

    print(count)
    print(count_ADC)



# coding: utf8

import clinica.engine as ce


class AiblToBidsCLI(ce.CmdParser):
    """
    todo:add description
    """

    def define_name(self):
        """Define the sub-command name to run this pipelines.
        """
        self._name = 'aibl-to-bids'

    def define_description(self):
        """Define a description of this pipeline.
        """
        self._description = 'Convert AIBL (https://aibl.csiro.au/adni/index.html”) into BIDS.'

    def define_options(self):
        """Define the sub-command arguments
        """
        self._args.add_argument("dataset_directory",
                                help='Path to the AIBL images directory.')
        self._args.add_argument("clinical_data_directory",
                                help='Path to the AIBL clinical data directory.')
        self._args.add_argument("bids_directory",
                                help='Path to the BIDS directory.')

    def run_command(self, args):
        from clinica.iotools.converters.aibl_to_bids.aibl_to_bids import convert_clinical_data, convert_images
        from clinica.utils.stream import cprint
        from os.path import exists
        from os import makedirs
        from colorama import Fore
        from clinica.iotools.converter_utils import check_bin
        import sys

        # Check existence of binaries dcm2nii, dcm2niix and mri_convert
        # If they are not found we warn the user and tell him that the bin
        # were not foud. He then has the possibility to run the converter anyway
        missing_bin = 0
        bin_to_test = ['dcm2nii', 'dcm2niix', 'mri_convert']
        for binary in bin_to_test:
            missing_bin = missing_bin + check_bin(binary)
        if missing_bin > 0:
            cprint(Fore.RED + str(missing_bin) + ' binary(es) is (are) missing. '
                   + 'Most important are : dcm2nii and dcm2niix.' + Fore.RESET)
            while True:
                cprint('Do you still want to run the converter ? (yes/no): ')
                answer = input('')
                if answer.lower() in ['yes', 'no']:
                    break
                else:
                    cprint('Possible answers are yes or no.\n')
            if answer.lower() == 'yes':
                cprint('Running the pipeline anyway.')
            if answer.lower() == 'no':
                cprint('Exiting clinica...')
                sys.exit()

        if not exists(args.bids_directory):
            makedirs(args.bids_directory)

        convert_images(args.dataset_directory, args.clinical_data_directory, args.bids_directory)
        convert_clinical_data(args.bids_directory, args.clinical_data_directory)
